from .admin import DataResource
from .base_models import *
from .correspondences import *
from .dims import *
from .external_schemas import *
from .facts import CountryFootprint, CountryRecipe, Footprint, Recipe
from .ipcc import ContentData, Parameters
from .matrix import A_Matrix, B_Matrix, IntensitiesMatrix, Inverse
from .metadata import DataLicense, MetaData, User, Version
from .PPF_fact_schemas_samples import *
from .PPF_fact_schemas_uncertainty import *
from .uncertainty import PedigreeMatrix, Uncertainty



import json
import logging
import os
from getpass import getpass
from pathlib import Path
from typing import Optional, Union

import pandas as pd
import requests
from dotenv import load_dotenv

from dataio.utils.schema_enums import APIEndpoints

load_dotenv()


# TODO: seperate cache logic out (@Fan 2024-10-24 13:52:30)
class Connector:
    def __init__(self, url, token: str = None):
        self.url = url
        # Check if the token was passed explicitly, otherwise try loading it from cache
        self.token = token

    def parse_url(self): ...

    def _set_header(self):

        if self.token:
            headers = {
                "Accept": "application/json",
                "Authorization": f"Token {self.token}",
            }
        else:
            headers = {
                "Accept": "application/json",
            }
        return headers

    def get(self):
        try:
            response = requests.get(self.url, headers=self._set_header())
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if response.status_code == 401:
                raise Exception(
                    f"Unauthorized: Your token {self.token} is not correctly setup."
                ) from e
            elif response.status_code == 429:
                raise Exception(
                    "Too many requests: Please wait and try again later."
                ) from e
        data = response.json()

        if "results" in data:
            return self._parse_paginated_response(data)
        else:
            return self._parse_non_paginated_response(data)

    def _parse_paginated_response(self, data):
        results = data["results"]
        while data.get("next"):
            response = requests.get(data["next"], headers=self._set_header())
            data = response.json()
            results.extend(data["results"])
        return results

    def _parse_non_paginated_response(self, data):
        return data

    def get_df(self) -> pd.DataFrame:
        return pd.DataFrame(self.get())

    def _pre_payload(
        self, data: Union[pd.DataFrame, pd.Series], chunk_size=None
    ) -> str:

        if isinstance(data, pd.DataFrame):
            if chunk_size is not None:
                # Creating chunks using list comprehension
                return [
                    data.iloc[i : i + chunk_size].to_json(orient="records")
                    for i in range(0, data.shape[0], chunk_size)
                ]
            else:
                return [data.to_json(orient="records")]
        # Check if the input data is a pandas Series
        elif isinstance(data, pd.Series):
            return [data.to_json(orient="index")]
        else:
            raise ValueError("Input data must be a pandas DataFrame or Series")

    def post(self, data: Union[pd.DataFrame, pd.Series], **kwargs):

        chunk_size = kwargs.pop("chunk_size", None)
        payloads = self._pre_payload(data, chunk_size=chunk_size)
        results = []
        for payload in payloads:
            payload_json = json.loads(payload)
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Token {self.token}",
            }
            response = requests.post(
                self.url, json=payload_json, headers=headers, **kwargs
            )  # Use remaining kwargs here
            if response.status_code == 201:
                results.append(response.json())
            else:
                raise Exception(
                    f"Post failed with status code {response.status_code}: {response.text}"
                )
        return results

    def put(self, data: Union[pd.DataFrame, pd.Series], **kwargs):
        chunk_size = kwargs.pop("chunk_size", None)
        payloads = self._pre_payload(data, chunk_size=chunk_size)
        results = []
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Token {self.token}",
        }

        for payload in payloads:
            payload_json = json.loads(payload)

            # If the payload is a list of records (from DataFrame)
            if isinstance(payload_json, list):
                for item in payload_json:
                    code = item.get("code")
                    if not code:
                        raise ValueError("Each item must contain a 'code' key.")
                    url_with_code = f"{self.url}/{code}"
                    response = requests.put(
                        url_with_code, json=item, headers=headers, **kwargs
                    )
                    if response.ok:
                        try:
                            results.append(response.json())
                        except ValueError:
                            results.append(None)
                    else:
                        raise Exception(
                            f"PUT failed with status code {response.status_code}: {response.text}"
                        )
            # If the payload is a single record (from Series)
            elif isinstance(payload_json, dict):
                code = payload_json.get("code")
                if not code:
                    raise ValueError("Data must contain a 'code' key.")
                url_with_code = f"{self.url}/{code}"
                response = requests.put(
                    url_with_code, json=payload_json, headers=headers, **kwargs
                )
                if response.ok:
                    try:
                        results.append(response.json())
                    except ValueError:
                        results.append(None)
                else:
                    raise Exception(
                        f"PUT failed with status code {response.status_code}: {response.text}"
                    )
            else:
                raise ValueError("Payload must be a dict or list of dicts.")

        return results


class ConnectorFactory:
    def __init__(self):
        self.cached_token = None

    def create(self, url: str):
        # TODO: token use configuration should be in Config class insted of ConnectorFactory (@Fan 2024-11-15 16:09:40)
        token = self._get_token()
        return Connector(url, token)

    def _get_token(self) -> str:
        """
        Retrieves a token from cache, environment, or prompts for credentials if none found.
        """
        if not self.cached_token:
            self.cached_token = self._load_token_from_cache() or os.environ.get(
                "BONSAI_TOKEN"
            )
            if not self.cached_token:

                # TODO: there should be a way to update token use setting (@Fan 2024-11-15 16:09:46)
                config = self._load_config()
                if not config["use_token"]:
                    return None
                else:
                    self.cached_token = self._prompt_for_credentials()
        return self.cached_token

    def _prompt_for_credentials(self) -> str:
        """
        Prompts the user for credentials and retrieves a token.
        """
        choice = (
            input(
                """
                No token found. Rate limiting is applied to API accessibility.\n
                Do you want to enter credentials to get a token to avoid rate limiting? (y/n): 
                """
            )
            .strip()
            .lower()
        )
        if choice == "y":
            email = input("Enter your email: ")
            password = getpass("Enter your password: ")
            token = self._get_token_from_api(email, password)
            if token:
                self._cache_token(token)
                self._save_user_decision(True)
                return token
            else:
                raise ValueError(
                    "Failed to retrieve token. Please check your credentials."
                )
        elif choice == "n":
            logging.warning(
                "Proceeding without a token. You may encounter rate limits."
            )
            self._save_user_decision(False)
            return ""
        else:
            logging.warning("Invalid choice. Please try again.")
            return self._prompt_for_credentials()

    def _get_token_from_api(self, email: str, password: str) -> str:
        """
        Retrieves a token by making a POST request to the token endpoint.
        """
        token_url = f"{APIEndpoints.BASE_URL.value}{APIEndpoints.TOKEN.value}"
        try:
            response = requests.post(
                token_url, data={"email": email, "password": password}
            )
            response.raise_for_status()
            token = response.json().get("token")
            if token:
                logging.info("Token successfully retrieved.")
                return token
        except requests.RequestException as e:
            logging.error(f"Error retrieving token: {e}")
        return ""

    def _load_token_from_cache(self) -> Optional[str]:
        """
        Loads the token from cache if it exists.
        """
        try:
            token_cache_path = self._get_cache_path()
            if os.path.exists(token_cache_path):
                with open(token_cache_path, "r") as f:
                    data = json.load(f)
                    token = data.get("token")
                    if token:
                        logging.info("Token loaded from cache.")
                        return token
        except IOError as e:
            logging.error(f"Failed to load token from cache: {e}")
        return None

    def _cache_token(self, token: str):
        """
        Saves the token to a cache file.
        """
        try:
            token_cache_path = self._get_cache_path()
            with open(token_cache_path, "w") as f:
                json.dump({"token": token}, f)
            logging.info(f"Token cached successfully at {token_cache_path}.")
        except IOError as e:
            logging.error(f"Failed to cache token: {e}")

    def _get_cache_path(self):
        """
        Returns the path for caching the token.
        """
        from pathlib import Path

        cache_path = Path.home() / ".cache" / "bonsai"
        cache_path.mkdir(parents=True, exist_ok=True)
        return cache_path / "token_cache.txt"

    def _load_config(self) -> dict:
        """
        Loads the user's choice from a configuration file.
        """
        config_path = self._get_config_path()
        if config_path.exists():
            try:
                with open(config_path, "r") as f:
                    return json.load(f)
            except (IOError, json.JSONDecodeError):
                logging.warning(
                    "Failed to load user decision. Defaulting to use token."
                )
        return {"use_token": True}  # Default to using tokens

    def _save_user_decision(self, use_token: bool):
        """
        Saves the user's choice to a configuration file.
        """
        config_path = self._get_config_path()
        try:
            with open(config_path, "w") as f:
                json.dump({"use_token": use_token}, f)
        except IOError as e:
            logging.error(f"Failed to save user decision: {e}")

    def _get_config_path(self) -> Path:
        """
        Returns the path for the configuration file.
        """
        config_dir = Path.home() / ".cache" / "bonsai"
        config_dir.mkdir(parents=True, exist_ok=True)
        return config_dir / "config.json"


class ConnectorRepository:
    def __init__(self, connector_factory: Optional[ConnectorFactory] = None):
        self.connector_factory = (
            connector_factory if connector_factory else ConnectorFactory()
        )
        self._cache = {}
        self._preload_connectors()

    def get(self, name: str) -> Connector:
        if name not in self._cache:
            url = APIEndpoints.get_url(name)
            self._cache[name] = self.connector_factory.create(url)
        return self._cache[name]

    def _preload_connectors(self):
        """
        Preload all connectors from APIEndpoints except BASE_URL.
        """
        for name in APIEndpoints.__members__:
            if name != "BASE_URL":  # Skip BASE_URL
                try:
                    self.get(name)  # This will cache the connector
                except ValueError as e:
                    logging.warning(f"Failed to preload connector for {name}: {e}")


connector_repository = ConnectorRepository()



"""Module with set_logger function.

Part of package 'templates'
Part of the Getting the Data Right project
Created on Jul 25, 2023
@author: Joao Rodrigues
"""

import logging
import os
import sys

# from datetime import datetime
from pathlib import Path
from typing import Union


def set_logger(
    filename: Union[str, Path] = None,
    path: str = os.getcwd(),
    log_level: int = 20,
    log_format: str = ("%(asctime)s | [%(levelname)s]: %(message)s"),
    overwrite=False,
    create_path=False,
) -> None:
    """Initialize the logger.

    This function creates and initializes a log file.
    Logging output is sent both to the file and standard output.
    If 'filename' == None, no file output is written
    To further write to this logger add in the script:

    import logging
    logger = logging.getLogger('root')
    logger.info(<info_string>)
    logger.warning(<warning_string>)
    logger.error(<error_string>)

    Parameters
    ----------
    filename : str
        name of output file
    path : str
        path to folder of output file
    log_level : int
        lowest log level to be reported. Options are:
          10=debug
          20=info
          30=warning
          40=error
          50=critical
    log_format : str
        format of the log
    overwrite : bool
      whether to overwrite existing log file
    create_path : bool
      whether to create path to log file if it does not exist
    """
    try:
        log_format_date = logging.Formatter(log_format, "%Y-%m-%dT%H:%M:%S")
    except ValueError:
        raise "Log format not allowed. Try using default instead."

    try:
        logging.basicConfig(level=log_level)
    except ValueError:
        raise "Log level not allowed. Try help(setlogger) for options."

    logger = logging.getLogger("root")
    if logger.hasHandlers():
        logger.handlers.clear()

    # setup writing to console
    handler_stdout = logging.StreamHandler(sys.stdout)
    handler_stdout.setLevel(log_level)
    handler_stdout.setFormatter(log_format_date)
    logger.addHandler(handler_stdout)

    # setup writing to file
    if filename is not None:
        logger.info("Not writing to file")

        try:
            full_path = Path(path).joinpath(filename)
        except TypeError:
            raise (
                f"Cannot combine '{path}' and '{filename}' "
                "in one path, are both types and values valid?"
            )

        if not os.path.exists(path):
            logger.info("Data output path does not exist")
            if create_path:
                os.makedirs(path)
                logger.info("Data output path created")
            else:
                logger.error(" and 'create_path' option is disabled")
                raise FileNotFoundError
        if os.path.exists(full_path):
            logger.info("Data output full path exists")
            if not overwrite:
                logger.error(" and 'overwrite' option is disabled")
                raise FileExistsError

        try:
            handler_file = logging.FileHandler(full_path, mode="w")
        except FileExistsError:
            raise (f"Error opening file path '{full_path}', does it exist?")
        handler_file.setLevel(log_level)
        handler_file.setFormatter(log_format_date)
        logger.addHandler(handler_file)

    # to avoid duplicates
    logger.propagate = False

    # To log uncaught exceptions when running as script
    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        logger.error(
            "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
        )

    sys.excepthook = handle_exception

    if filename is not None:
        logger.info(f"Logger initialized at {full_path}")
    else:
        logger.info(f"Logger initialized at console only")



from __future__ import annotations

import json
import logging
import os
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, fields
from datetime import datetime
from pathlib import Path
from typing import List, Union

from dataio.utils.accounts import AccountRepository
from dataio.utils.schema_enums import APIEndpoints


@dataclass
class Version:
    name: str
    acronym: str
    version_date: str  # Default Date in YYYY-MM-DD format
    version_semantic: str = None
    is_latest: bool = True

    def has_valid_version_date_format(self, expected_format: str = "%Y-%m-%d") -> bool:
        """
        Checks if the version_date is in the expected format.
        """
        try:
            datetime.strptime(self.version_date, expected_format)
            return True
        except ValueError:
            logging.warning(
                f"The current version date is in {self.version_date}. Please consider using recommended format: {expected_format}"
            )
            return False

    def parse_date(self, from_format: str = "%Y-%m-%d", to_format: str = "%Y%m%d"):
        """
        Parses the version_date from 'YYYY-MM-DD' to 'YYYYMMDD' format.
        """
        print(f"parsing date from {from_format} to {to_format}")
        parsed_date = datetime.strptime(self.version_date, from_format).strftime(
            to_format
        )
        self.version_date = parsed_date
        return self

    def update_version_date(self, version_date: str):
        """
        Updates the version_date to a new date provided in 'YYYY-MM-DD' format.
        """
        # Check if the new date is in the correct format
        try:
            date = datetime.strptime(version_date, "%Y-%m-%d").strftime("%Y-%m-%d")
            self.version_date = date
        except ValueError:
            print(f"Unexpected date format: {version_date}. Expected format: %Y-%m-%d")
        return self

    def to_json(self):
        data = asdict(self)

        return json.dumps(data)


class VersionCollection:
    def __init__(
        self,
        versions: List[Version],
        source: str | Path = None,
        account_repo: AccountRepository = None,
    ):
        self.versions = versions
        self.source = source
        default_account = (
            Path(os.environ.get("BONSAI_HOME")) / "_bonsai" / "accounts.json"
        )
        self.account_repo = account_repo or AccountRepository(default_account)

    @classmethod
    def load_from(cls, source: Union[Path, str]):
        if isinstance(source, str):
            versions = APIVersionSource(url=source).load()
        elif isinstance(source, Path):
            versions = TXTVersionSource(path=source).load()
        else:
            raise TypeError(f"version from {source} is not supported.")

        if not versions:
            raise ValueError(f"No versions found from neither {str(source)} nor API")
        return versions

    def get_account_versions(self, account: str) -> List[Version]:
        account_versions = [v for v in self.versions if v.name == account]
        if not account_versions:
            raise ValueError(f"No version found for account {account}")

        return account_versions

    def update_account_version_date(self, account_name: str, version_date: str):
        """Replace the version_date of account in version collection with a new version_date"""
        account = self.account_repo.get_or_create_account(account_name)
        try:
            account_version = self.get_latest_version(account.name)
            self.versions.remove(account_version)
            account_version.update_version_date(version_date)
            self.versions.append(account_version)
            self.save()
        except ValueError:
            print(f"Adding a new version record for {account.name}.")
            self.add_account_version(
                account_name=account_name, version_date=version_date
            )

    def get_latest_version(self, account: str) -> Version:
        account_versions = self.get_account_versions(account)

        latest_version = max(account_versions, key=lambda x: x.version_date)
        if not latest_version:
            raise ValueError(
                f"No version is assigned with the latest version for account {account}"
            )
        return latest_version

    def add_account_version(
        self,
        account_name: str,
        version_date: str,
        is_latest: bool = True,
        version_semantic: str = None,
    ) -> Version:
        """
        Adds a new version for an account. If the account does not exist, it adds the account.
        """
        # Ensure the account exists in the AccountRepository
        account = self.account_repo.get_or_create_account(account_name)

        # Check if a version with the same date already exists to prevent duplicates
        try:
            self.get_account_versions(account.name)
        except ValueError:
            print(
                f"No version found for account {account.name}, will add a new version using {version_date} for the account"
            )
        # Create a new version instance
        new_version = Version(
            name=account.name,
            acronym=account.abbreviation,
            version_date=version_date,
            version_semantic=version_semantic,
            is_latest=is_latest,
        )

        self.versions.append(new_version)
        # Optionally save the updated versions list
        self.save()

        return new_version

    def to_api(self, url: str = "https://lca.aau.dk/api/datasets/"):
        APIVersionSource(url).save()

    def to_txt(self, path: Path):
        TXTVersionSource(path=path).save(self.versions)

    def save(self):
        if isinstance(self.source, Path):
            self.to_txt(self.source)
        elif isinstance(self.source, str):
            self.to_api(self.source)
        else:
            raise ValueError("Unknown version source type.")


class VersionSource(ABC):
    @abstractmethod
    def load(self) -> VersionCollection:
        raise NotImplementedError

    @abstractmethod
    def save(self, versions: list[Version]):
        raise NotImplementedError


class APIVersionSource:
    def __init__(self, name: str = APIEndpoints.METADATA.name) -> None:
        self.name = name
        from dataio.utils.connectors import connector_repository

        self.api_connector = connector_repository.get(name=self.name)

    def load(self) -> VersionCollection:
        api_response = self.api_connector.get()
        version_fields = {field.name for field in fields(Version)}
        versions = []
        for metadata in api_response:
            version_data = {k: v for k, v in metadata.items() if k in version_fields}
            version_instance = Version(**version_data)
            version_instance.parse_date()
            versions.append(version_instance)
        if not versions:
            raise ValueError(f"No version found from {self.url}")
        return VersionCollection(versions, source=self.url)

    def save(self, versions: list[Version]):

        ls_version = []
        for version in versions:
            if not version.has_valid_version_date_format():
                version.parse_date(from_format="%Y%m%d", to_format="%Y-%m-%d").to_json()
            ls_version.append(version.to_json())

        return self.api_connector.post(ls_version)


class TXTVersionSource:
    def __init__(self, path: Path) -> None:
        self.path = path

    def load(self):
        """
        read the version-date of a previously produced dataset
            Here the list of datasets (in key) and the acronym
            inserted in the version.txt:

        parameters
        --------------
        path: path of the version file.

        """
        "dictionary to link datasets to acronyms used in version file"
        default_account = (
            Path(os.environ.get("BONSAI_HOME")) / "_bonsai" / "accounts.json"
        )
        account_repo = AccountRepository(default_account)

        with open(self.path, "r") as file:
            version_txt = file.read()

        versions = []
        for account in account_repo.accounts.values():
            # position where the date to read starts
            start_change = version_txt.find(account.abbreviation) + 4
            if start_change < 4:
                logging.info(
                    f"Acronym {account.abbreviation} for account {account.name} not found in the source: {str(self.path)}."
                )
                continue

            # date to read
            date_to_read = version_txt[start_change : start_change + 10]

            account_version = Version(
                name=account.name,
                acronym=account.abbreviation,
                version_date=date_to_read,
            )
            if not account_version.has_valid_version_date_format():
                account_version.parse_date(from_format="%Y%m%d", to_format="%Y-%m-%d")
                logging.info(f"parsing date for {account} successful")
            versions.append(account_version)
        return VersionCollection(versions, source=self.path)

    def save(self, versions: list[Version]):
        with open(self.path, "w") as file:
            for version in versions:
                if not version.has_valid_version_date_format():
                    version.parse_date(from_format="%Y%m%d", to_format="%Y-%m-%d")
                file.write(f"{version.acronym}='{version.version_date}'" + "\n")



from dataclasses import dataclass
from enum import Enum


class IndexNames(Enum):
    OJBECT_CODE = "object_code"
    PROD_CODE = "Exio prod code"
    ACT_CODE = "Exio act code"
    COUNTRY_CODE = "Exio country code"
    AGGREGATION = "Aggregation"
    SUBSTITUTION_FACTOR = "substitution factor"
    REPLACED_PRODUCT = "replaced product"
    REPL_FACTOR = "repl factor"
    PRODUCT = "product"
    PRODUCTION = "production"
    UNIT = "unit"
    VALUE = "Value"
    MARKET = "market"
    FLAG = "flag"
    ORIGIN_COUNTRY = "Exio origin country code"
    DESTIN_COUNTRY = "Exio destin. country code"
    FACTOR = "factor"
    SHARE = "share"
    EXIO3 = "nace-related code"
    EMIS_SUBSTANCE = "Exio substance"
    EMIS_COMPARTMENT = "Exio compartment"
    GLOBAL_AREA = "Global"
    RESOURCE = "Exio resource"
    EXIO_ACT = "Exiobase activity"  # TODO: used at the beginning in IEA when the workflow was not clear, to be replaced with "Exio act code"
    EXIO_CNT = "Exiobase country"
    EXIO_CNT_acron = "Exiobase cnt"
    EXIO_PRD = "Exiobase product"
    EXIO_CODE = "Exiobase code"
    DESCRIP = "description"
    REPLACED_MRK = "Replaced market"
    POSITION = "position"
    PACK_CODE = "exio4_pack_id"
    WASTE_FRACTION = "Exio waste"
    WASTE_MARKET = "Exio waste market"
    PACK_PROD = "Exio packaging prod code"
    PACKAGED = "Packaged product"
    PACK_MARKET = "Exio packaging market"
    INPUT_PROD = "Exio prod code (input)"
    VALUE_IN = "Value (input)"
    VALUE_FOOTPRINT = "Value (footprint)"
    UNIT_DESTIN = "unit (source)"
    UNIT_SOURCE = "unit (input)"
    UNIT_FOOTPRINT = "unit (footprint)"
    CLIMATE_METRIC = "climate metric"
    ANIMAL_CATEGORY = "animal categ"
    PERIOD = "time"
    AGRI_SYSTEM = "system"
    EXIO_ACCOUNT = "Account"
    NUTRIENT_CONT = "nutrient content"
    COEFFICIENT = "coeff"
    REF_PROD_CODE = "Exio ref. prod code"
    LCI_VALUE = "Value_lci"
    LCI_UNIT = "unit_lci"
    LCI_FLAG = "flag_lci"
    SOURCE = "source"
    SOURCE_CODE = "source_code"
    SOURCE_NAME = "source_name"
    SOURCE_CLASS = "source_class"
    SOURCE_LINK = "source_link"
    TARGET_CODE = "target_code"
    TARGET_NAME = "target_name"
    TARGET_CLASS = "target_class"
    TARGET_LINK = "target_link"
    PARENT_CODE = "parent_code"
    PARENT_NAME = "parent_name"
    PARENT_CLASS = "parent_class"
    PARENT_LINK = "parent_link"
    SCENARIO = "scenario"
    TRADE_ROUTE_ID = "trade route ID"
    TRANSPORT_MODE = "transport mode"
    EDGE = "edge"
    ORIGIN_COUNTRY_EDGE = "Exio origin country code edge"
    DESTIN_COUNTRY_EDGE = "Exio destin. country code edge"
    TRANSPORT_MODE_EDGE = "transport mode edge"
    EXIO3_ACT = "exio_v3"
    PERIOD_DELAY = "time delay"
    ASSOCIATED_TREAT = "associated treatment service"


class EmissCompartment(Enum):
    AIR = "Air"
    WATER = "Water"
    SOIL = "Soil"


class land_use_categ(Enum):
    sf_to_cp = "secondary forest to crop"
    pf_to_mf = "primary forest to managed forest"
    sf_to_mf = "secondary forest to managed forest"
    gr_to_ps = "grassland to pasture"
    int_cp = "intensification of cropland"
    int_ps = "intensification of pasture"
    unit_land = "ha-weighted"  # TODO: to be replaced by PropertyEnum.LAND.unit


class global_land_categ(Enum):
    arable = "Arable Land"
    forest = "Forest Land"
    grass = "Grassland"


class fao_categ(Enum):
    ITEM = "Item"
    ELEM = "Element"
    UNIT = "Unit"
    AREA = "Area"


class animal_system(Enum):
    MILK = "milk"
    MEAT = "meat"


class ipcc_categ(Enum):
    ef4_wet = "Wet climate"
    ef4_dry = "Dry climate"
    fract_leach_wet = "Wet climate"


@dataclass
class Property:
    name: str
    value: float
    unit: str
    upperbound: float | None = None
    lowerbound: float | None = None


class PropertyEnum(Enum):
    MASS = ("tonnes", Property)
    ENERGY = ("TJ", Property)
    CURRENCY = ("Meuro", Property)
    ITEM_COUNT = ("items", Property)
    WASTE_SERVICE = ("tonnes (service)", Property)
    LAND = ("ha-weighted", Property)
    FREIGHT = ("tkm", Property)
    NA = ("undefined", Property)

    def __init__(self, unit: str, cls: Property) -> None:
        self.unit = unit
        self.cls = cls

    def create_property(
        self, value, upperbound=float("inf"), lowerbound=float("-inf")
    ) -> Property:
        return self.cls(
            value=value,
            unit=self.unit,
            name=self.name,
            upperbound=upperbound,
            lowerbound=lowerbound,
        )


# TODO: Use service discovery tools when we get more endpoints. (@Fan)
class APIEndpoints(Enum):
    BASE_URL = "https://lca.aau.dk/api/"
    ACTIVITIES = "activity-names/"
    ACTIVITY_CORR = "activity-corr/"
    PRODUCTS = "products/"
    PRODUCT = "product/"  # TODO: use plural form (@Fan 2024-10-03 15:15:15)
    PRODUCT_CORR = "product-corr/"
    LOCATIONS = "locations/"
    LOCATION_CORR = "location-corr/"
    METADATA = "datasets/"
    FOOTPRINT = "footprint/"
    PROPERTIES = "properties/"
    SUPPLY = "supply/"
    USE = "use/"
    RECIPES = "recipes/"
    TOKEN = "user/token/"

    @classmethod
    def get_url(cls, name: str) -> str:
        """
        Return the full URL for a given endpoint name.

        Parameters:
        - name: The name of the endpoint to retrieve.

        Returns:
        - The full URL of the requested endpoint.
        """
        # Ensure the requested name is a valid enum member
        if name in cls.__members__:
            endpoint = cls[name].value
            # Special handling for BASE_URL
            if name == "BASE_URL":
                return endpoint
            return f"{cls.BASE_URL.value}{endpoint}"
        else:
            raise ValueError(f"{name} is not a valid API endpoint")


class Exio_fert_nutrients:
    nutrients = ["N", "P2O5", "K2O"]
    market_dict = {"N": "M_Nfert", "P2O5": "M_P2O5fert", "K2O": "M_K2Ofert"}


@dataclass
class data_index_categ:
    # data structure of emissions df
    emiss_categ = [
        IndexNames.COUNTRY_CODE.value,
        IndexNames.EMIS_SUBSTANCE.value,
        IndexNames.EMIS_COMPARTMENT.value,
        IndexNames.ACT_CODE.value,
        IndexNames.UNIT.value,
    ]

    # data  use and supply
    general_categ = [
        IndexNames.ORIGIN_COUNTRY.value,
        IndexNames.PROD_CODE.value,
        IndexNames.DESTIN_COUNTRY.value,
        IndexNames.ACT_CODE.value,
        IndexNames.UNIT.value,
    ]

    # data  use and supply
    trade_categ = [
        IndexNames.ORIGIN_COUNTRY.value,
        IndexNames.PROD_CODE.value,
        IndexNames.DESTIN_COUNTRY.value,
        IndexNames.UNIT.value,
    ]

    balance_categ = [
        IndexNames.ORIGIN_COUNTRY.value,
        IndexNames.PROD_CODE.value,
        IndexNames.DESTIN_COUNTRY.value,
        IndexNames.ACT_CODE.value,
    ]

    balance_columns = [
        IndexNames.VALUE.value,
        IndexNames.FLAG.value,
        IndexNames.UNIT.value,
    ]

    # data  use and supply
    column_categ = [IndexNames.VALUE.value, IndexNames.FLAG.value]

    pack_index = [
        IndexNames.ORIGIN_COUNTRY.value,
        IndexNames.PROD_CODE.value,
        IndexNames.DESTIN_COUNTRY.value,
        IndexNames.ACT_CODE.value,
        IndexNames.PACK_CODE.value,
    ]

    waste_sup_index = [
        IndexNames.ORIGIN_COUNTRY.value,
        IndexNames.PROD_CODE.value,
        IndexNames.DESTIN_COUNTRY.value,
        IndexNames.ACT_CODE.value,
        IndexNames.WASTE_FRACTION.value,
    ]

    waste_sup_col = [
        IndexNames.VALUE.value,
        IndexNames.FLAG.value,
        IndexNames.UNIT.value,
    ]

    fao_clean_index = [
        IndexNames.COUNTRY_CODE.value,
        fao_categ.ITEM.value,
        fao_categ.ELEM.value,
        fao_categ.UNIT.value,
    ]

    fao_animal_system_index = [
        IndexNames.COUNTRY_CODE.value,
        fao_categ.ITEM.value,
        fao_categ.ELEM.value,
        IndexNames.AGRI_SYSTEM.value,
        IndexNames.ANIMAL_CATEGORY.value,
        IndexNames.UNIT.value,
        IndexNames.PERIOD.value,
    ]



import pandas as pd
import pytest

from dataio._classifications_helper import *


@pytest.fixture
def setup_classification_data():
    # Example classification data setup
    data = [
        [
            "C_ANTH",
            "0100",
            "many-to-many correspondence",
            "9e3aea85-6ef1-49ab-98df-e0bdf80644af",
            "http://www.w3.org/2004/02/skos/core#relatedMatch",
        ],
        [
            "C_COKC",
            "0100",
            "many-to-many correspondence",
            "2e4de738-b168-46d8-9132-086586cd44d5",
            "http://www.w3.org/2004/02/skos/core#relatedMatch",
        ],
        [
            "C_ANTH",
            "0110",
            "one-to-many correspondence",
            "f7c57ebe-1027-47ea-ab76-27058ae87a28",
            "http://www.w3.org/2004/02/skos/core#closeMatch",
        ],
        [
            "C_COKC",
            "0121",
            "one-to-many correspondence",
            "c4f4f717-0556-448c-8226-7d247e684e6d",
            "http://www.w3.org/2004/02/skos/core#closeMatch",
        ],
        [
            "C_OTBC",
            "0129",
            "one-to-many correspondence",
            "e2e7da7f-00a1-411c-88f9-faaae7b89906",
            "http://www.w3.org/2004/02/skos/core#closeMatch",
        ],
    ]

    columns = [
        "bonsai_classification",
        "source_classification",
        "comment",
        "uuid",
        "relation",
    ]
    classification = pd.DataFrame(data, columns=columns)

    return classification


def test_filter_classifications(setup_classification_data):

    classification = setup_classification_data

    data = pd.DataFrame(
        [["a", "0121"], ["0121", "0129"], ["0121", "0130"]], columns=["one", "two"]
    )

    filtered_table = filter_classifications(
        classification, data["two"].unique(), "source_classification"
    )
    assert len(filtered_table) == 2


def test_filter_classifications_with_regex():
    classification_data = [
        [
            "C_ANTH",
            "0100",
            "many-to-many correspondence",
            "9e3aea85-6ef1-49ab-98df-e0bdf80644af",
            "http://www.w3.org/2004/02/skos/core#relatedMatch",
        ],
        [
            "C_COKC",
            "0100",
            "many-to-many correspondence",
            "2e4de738-b168-46d8-9132-086586cd44d5",
            "http://www.w3.org/2004/02/skos/core#relatedMatch",
        ],
        [
            "C_ANTH",
            "0110",
            "one-to-many correspondence",
            "f7c57ebe-1027-47ea-ab76-27058ae87a28",
            "http://www.w3.org/2004/02/skos/core#closeMatch",
        ],
        [
            "C_COKC",
            "Co[k]?e",
            "one-to-many correspondence",
            "c4f4f717-0556-448c-8226-7d247e684e6d",
            "http://www.w3.org/2004/02/skos/core#closeMatch",
        ],
        [
            "C_OTBC",
            "0129",
            "one-to-many correspondence",
            "e2e7da7f-00a1-411c-88f9-faaae7b89906",
            "http://www.w3.org/2004/02/skos/core#closeMatch",
        ],
    ]

    columns = [
        "bonsai_classification",
        "source_classification",
        "comment",
        "uuid",
        "relation",
    ]
    classification = pd.DataFrame(classification_data, columns=columns)

    data = pd.DataFrame(
        [["a", "Coke"], ["0121", "no_match"], ["0121", "also_no_match"]],
        columns=["one", "two"],
    )

    filtered_table = filter_classifications(
        classification, data["two"].unique(), "source_classification", using_regex=True
    )
    assert len(filtered_table) == 1


def test_find_nearest_parent():
    bonsai_codes = ["A", "B", "C"]
    tree_bonsai_df = pd.DataFrame(
        {
            "code": ["A", "B", "C", "D", "E"],
            "parent_code": [None, "A", "A", "B", "B"],
            "level": [0, 1, 1, 2, 2],
        }
    )

    common_parent, max_steps = find_nearest_parent(bonsai_codes, tree_bonsai_df)

    assert common_parent == "A"
    assert max_steps == 1


def test_combine_duplicates():
    # Test data
    data = {
        "time": ["2024-08-16", "2024-08-16", "2024-08-17", "2024-08-16"],
        "location": ["Store1", "Store1", "Store2", "Store1"],
        "product": ["ProductA", "ProductA", "ProductB", "ProductA"],
        "unit": ["Box", "Box", "Box", "Box"],
        "value": [10, 15, 5, 20],
        "extra_info": ["Info1", "Info2", "Info3", "Info4"],  # Additional column
    }

    df = pd.DataFrame(data)

    # Expected result after combining duplicates
    expected_data = {
        "time": ["2024-08-16", "2024-08-17"],
        "location": ["Store1", "Store2"],
        "product": ["ProductA", "ProductB"],
        "unit": ["Box", "Box"],
        "value": [45, 5],  # 10 + 15 + 20 = 45
        "extra_info": ["Info1", "Info3"],  # Keep first occurrence
    }
    expected_df = pd.DataFrame(expected_data)

    # Run the combine_duplicates function
    result_df = combine_duplicates(df)

    # Test if the result matches the expected output
    assert result_df.equals(
        expected_df
    ), "Test failed: The combined DataFrame does not match the expected output."


def test_increment_version_patch():
    assert increment_version("1.0.0") == "1.0.1"
    assert increment_version("2.5.9") == "2.5.10"
    assert increment_version("0.0.9") == "0.0.10"
    assert increment_version("3.9.99") == "3.9.100"


def test_increment_version_minor():
    assert increment_version("1.0") == "1.1"
    assert increment_version("2.5") == "2.6"


def test_increment_version_major():
    assert increment_version("1") == "2"
    assert increment_version("9") == "10"


def test_increment_version_invalid_format():
    # Test case for empty string
    try:
        increment_version("")
        assert False, "Expected ValueError for empty string"
    except ValueError:
        pass  # Expected outcome

    # Test case for non-numeric values
    try:
        increment_version("abc.def.ghi")
        assert False, "Expected ValueError for non-numeric input"
    except ValueError:
        pass  # Expected outcome



from collections import defaultdict

import pandas as pd


def filter_classifications(
    correspondence, unique_values, origin_column, using_regex=False
) -> dict[str, str]:
    """
    returns a filtered correspondence without many-to-many relations
    this is needed as many-to-many relations cannot be resolved
    automatically
    This also makes sure that only relavent correspondences are used
    filtered by the unique names.
    If using_regex is True then the function will make sure to already,
    parse potential fields and create a custom correspondence, this is done
    in order to not have to do regex matching when applying the mapping to the
    full dataset.
    """
    # change correspondence for regex values
    if using_regex:
        import re

        mapping_for_matched_regex = {}
        for regex in correspondence[origin_column]:
            for unique_value in unique_values:
                if re.findall(regex, unique_value):
                    assert (
                        regex not in mapping_for_matched_regex
                    ), f"Classification regex {regex} matched more then a single unique value, this should not be possible. Please revise classification regex."
                    mapping_for_matched_regex[regex] = unique_value

        correspondence[origin_column] = correspondence[origin_column].apply(
            lambda x: (
                mapping_for_matched_regex[x]
                if x in mapping_for_matched_regex.keys()
                else x
            )
        )

    # Make sure that only actually used unique values are in the correspondence
    correspondence = correspondence[correspondence[origin_column].isin(unique_values)]

    # Identify the many-to-many correspondences
    many_to_many = correspondence[
        correspondence["comment"] == "many-to-many correspondence"
    ]

    # Create a set of classifications involved in many-to-many relationships
    involved_classifications = set(many_to_many.iloc[:, 0]).union(
        set(many_to_many.iloc[:, 1])
    )

    # Identify rows involved in many-to-many relationships
    relations_to_ignore = correspondence[
        correspondence.iloc[:, 0].isin(involved_classifications)
        | correspondence.iloc[:, 1].isin(involved_classifications)
    ]

    # Create the filtered table by excluding the rows involved in many-to-many relationships
    filtered_table = correspondence[
        ~correspondence.index.isin(relations_to_ignore.index)
    ]

    return filtered_table


def generate_classification_mapping_multi_column(
    filtered_table, target_columns
) -> tuple[dict[tuple, (tuple, str)], set[tuple]]:
    mapping = {}
    # accounttype = []
    many_to_one = defaultdict(list)  # Reverse lookup for many-to-one

    for _, row in filtered_table.iterrows():
        to_tuple = tuple(row[f"{col}_to"] for col in target_columns)
        from_tuple = tuple(row[f"{col}_from"] for col in target_columns)
        comment = row["comment"]
        # accounttype = row["accounttype"]

        if comment == "one-to-one correspondence":
            mapping[from_tuple] = to_tuple

        elif comment == "many-to-one correspondence":
            mapping[from_tuple] = to_tuple
            many_to_one[to_tuple].append(from_tuple)

        elif comment == "one-to-many correspondence":
            if from_tuple in mapping:
                # Extend existing tuple by joining new values with '|'
                mapping[from_tuple] = tuple(
                    f"{mapping[from_tuple][i]}|{to_tuple[i]}"
                    for i in range(len(to_tuple))
                )
            else:
                mapping[from_tuple] = to_tuple

    return mapping, many_to_one


def generate_classification_mapping(
    filtered_table, target_column
) -> tuple[dict[str, str], set[str]]:
    mapping = {}
    from_col = target_column + "_from"
    to_col = target_column + "_to"

    # Collect many-to-one correspondences into a set
    many_to_one = set(
        filtered_table.loc[
            filtered_table["comment"] == "many-to-one correspondence", to_col
        ]
    )

    # Find one-to-many correspondences and join the codes in 'to_col'
    one_to_many_mask = filtered_table["comment"] == "one-to-many correspondence"
    filtered_table.loc[one_to_many_mask, to_col] = (
        filtered_table.loc[one_to_many_mask]
        .groupby(from_col)[to_col]
        .transform(lambda x: "|".join(x))
    )
    # Combine the from_col and to_col into a dictionary
    mapping = dict(zip(filtered_table[from_col], filtered_table[to_col]))

    return mapping, many_to_one


def find_nearest_parent(bonsai_codes, tree_bonsai_df):
    tree = build_tree(tree_bonsai_df)
    paths = {code: find_path_to_root(code, tree) for code in bonsai_codes}

    common_parent = None
    min_steps = float("inf")

    for i, (code1, path1) in enumerate(paths.items()):
        for j, (code2, path2) in enumerate(paths.items()):
            if i >= j:
                continue
            common, steps = find_common_ancestor_and_steps(path1, path2)
            if steps < min_steps:
                common_parent = common
                min_steps = steps

    # Calculate the maximum number of steps to the common ancestor for all paths
    max_steps = 0
    for path in paths.values():
        steps_to_common = find_steps_to_common_ancestor_using_levels(
            path, tree_bonsai_df, common_parent
        )
        if steps_to_common > max_steps:
            max_steps = steps_to_common

    return common_parent, max_steps


def build_tree(df):
    tree = {}
    for _, row in df.iterrows():
        tree[row["code"]] = row["parent_code"]
    return tree


def find_path_to_root(code, tree):
    path = []
    while code:
        path.append(code)
        code = tree.get(code)
    return path[::-1]


def find_common_ancestor_and_steps(path1, path2):
    common_ancestor = None
    steps = 0
    for code1, code2 in zip(path1, path2):
        if code1 == code2:
            common_ancestor = code1
        else:
            break
        steps += 1
    return common_ancestor, steps


def find_steps_to_common_ancestor_using_levels(path, tree_bonsai_df, common_ancestor):
    # Find the relevant levels in the DataFrame
    levels = tree_bonsai_df.loc[tree_bonsai_df["code"].isin(path)]["level"]

    # Check if levels is empty
    if levels.empty:
        return 0

    max_level = max(levels)
    min_level = min(levels)

    return max_level - min_level


def combine_duplicates(df):
    """
    Combine duplicate rows in a DataFrame by summing up their 'value' column, treating None values as 0.

    Parameters:
    df (pd.DataFrame): Input DataFrame with potential duplicate rows.

    Returns:
    pd.DataFrame: Combined DataFrame with duplicates merged.
    """
    # Define the columns to check for duplicates
    overlap_columns = ["time", "location", "product", "unit"]

    # Define the value column that needs to be summed
    value_column = "value"

    # Ensure that the value column is in the DataFrame
    if value_column not in df.columns:
        raise ValueError(f"The DataFrame does not contain a '{value_column}' column.")

    # Convert the value column to numeric, coercing errors to NaN
    df[value_column] = pd.to_numeric(df[value_column], errors="coerce").fillna(
        0.0
    )  # Replace NaNs with 0.0

    # Group by the overlap columns and aggregate using sum for the value column
    combined_df = df.groupby(overlap_columns, as_index=False).agg(
        {
            value_column: "sum",
            **{
                col: "first"
                for col in df.columns
                if col not in overlap_columns and col != value_column
            },
        }
    )

    return combined_df


def count_rows_with_values(
    df: pd.DataFrame, column_name: str, values_list: list
) -> int:
    """
    Counts the number of rows in a DataFrame where the specified column contains any of the given values.

    Parameters:
    - df (pd.DataFrame): The DataFrame to search within.
    - column_name (str): The name of the column to check.
    - values_list (list): The list of values to look for in the specified column.

    Returns:
    - int: The count of rows that have any of the specified values in the given column.
    """
    if column_name not in df.columns:
        raise ValueError(f"Column '{column_name}' does not exist in the DataFrame.")

    if not isinstance(values_list, list):
        raise TypeError("The 'values_list' parameter should be a list.")

    # Count rows where the column has values in values_list
    count = df[df[column_name].isin(values_list)].shape[0]
    return count


def increment_version(version):
    # Split the version string into its components
    version_parts = version.split(".")

    # Convert the last part to an integer and increment it
    version_parts[-1] = str(int(version_parts[-1]) + 1)

    # Reassemble the parts back into a string
    new_version = ".".join(version_parts)

    return new_version



import inspect
import logging
import os
from datetime import datetime
from pathlib import Path, PurePath, PureWindowsPath
from typing import Optional, Union
from uuid import UUID

import pandas as pd
from pydantic import ConfigDict, Field, computed_field

from dataio.tools import BonsaiBaseModel


class DataResource(BonsaiBaseModel):
    # model_config = ConfigDict(populate_by_name=True)

    id: UUID | None = None  # Only used for api
    run_by_user: str | None = None
    name: str
    schema_name: str
    task_name: str | None = None
    stage: str | None = None
    data_flow_direction: str | None = None
    data_version: str | None = None
    code_version: str | None = None
    comment: str | None = None
    last_update: datetime = Field(default_factory=datetime.now)
    created_by: str | None = None  # Assuming a ForeignKey-like relationship
    license: str | None = None
    license_url: str | None = None
    dag_run_id: str | None = None
    url: str | None = None
    _location: str = "{stage}/{task_name}/{version}/{resource_name}.csv"
    _root_location: str | None = None
    api_endpoint: str | None = None

    @computed_field
    @property
    def location(self) -> str:
        location = self._replace_variables_in_location_string(
            self._location,
            self.name,
            self.task_name,
            self.stage,
            self.data_version,
            self.code_version,
        )
        return self._parse_location(location, self._root_location)

    @location.setter
    def location(self, value: str | PurePath):
        if isinstance(value, str):
            if not value.startswith("http"):
                value = PureWindowsPath(value).as_posix()
        else:
            value = value.as_posix()
        self._location = value

    def __str__(self) -> str:
        return self.name

    def __init__(
        self,
        name: str,
        schema_name: str | BonsaiBaseModel,
        location: str | Path | None = None,
        task_name: str | None = None,
        stage: str | None = None,
        data_flow_direction: str | None = None,
        data_version: str | float | None = None,
        code_version: str | float | None = None,
        comment: str | None = None,
        last_update: datetime | None = None,
        created_by: str | None = None,  # Assuming a ForeignKey-like relationship
        license: str | None = None,
        license_url: str | None = None,
        dag_run_id: str | None = None,
        url: str | None = None,
        root_location: str | Path | None = None,
        api_endpoint: str | None = None,
        id: Optional[UUID] = None,
        run_by_user: str | None = None,
    ) -> None:
        # TODO potentially also allow to automatically generate api endpoint

        if inspect.isclass(schema_name):
            schema_name = schema_name.__name__

        # Dictionary to hold arguments that are not None
        init_args = {"name": name, "schema_name": schema_name}

        # Optional parameters are added only if they are not None
        if isinstance(id, str) and not id:
            # If id is an empty string, treat it as None
            id = None
        if id is not None:
            init_args["id"] = id
        if task_name is not None:
            init_args["task_name"] = task_name
        if stage is not None:
            init_args["stage"] = stage
        if data_flow_direction is not None:
            init_args["data_flow_direction"] = data_flow_direction
        if data_version is not None:
            init_args["data_version"] = data_version
        if code_version is not None:
            init_args["code_version"] = code_version
        if comment is not None:
            init_args["comment"] = comment
        if last_update is not None:
            init_args["last_update"] = last_update
        if created_by is not None:
            init_args["created_by"] = created_by
        if license is not None:
            init_args["license"] = license
        if license_url is not None:
            init_args["license_url"] = license_url
        if url is not None:
            init_args["url"] = url
        if api_endpoint is not None:
            init_args["api_endpoint"] = api_endpoint
        if isinstance(run_by_user, str) and not run_by_user:
            # If id is an empty string, treat it as None
            run_by_user = None
        if run_by_user is not None:
            init_args["run_by_user"] = run_by_user

        # Call to the superclass constructor with filtered parameters
        super().__init__(**init_args)

        # making sure that all paths are always handled as posixpaths
        if location is not None:
            self.location = location

        if root_location:
            if isinstance(root_location, str):
                if root_location.startswith("http"):
                    self._root_location = root_location
                else:
                    self._root_location = PureWindowsPath(root_location).as_posix()
            else:
                self._root_location = PureWindowsPath(root_location).as_posix()

    @staticmethod
    def _parse_location(location: str, root_location: str | None = None) -> str:
        assert location is not None, "Location can't be None"

        if not location.startswith("http"):

            # Check if loc is an absolute path
            path = Path(location)
            if path.is_absolute():
                # If the location is absolute, we assume that this is on purpose
                # and that DATAIO_ROOT should not be used.
                # If it is not absolute, we assume that BONSAI_HOME is the root
                # folder.
                return str(path)
            else:
                if root_location:
                    return str(root_location / path)
                else:
                    bonsai_home = os.getenv("DATAIO_ROOT", None)
                    assert bonsai_home, EnvironmentError(
                        "DATAIO_ROOT environmental varaible is not set, this needs to be set if you are not using an API location."
                    )
                    bonsai_home = Path(bonsai_home)
                    return str(bonsai_home / path)
        else:
            return location

    @staticmethod
    def _replace_variables_in_location_string(
        location: str,
        name=None,
        task_name=None,
        stage=None,
        data_version=None,
        code_version=None,
    ):
        if "{name}" in location or "{resource_name}" in location:
            assert (
                name
            ), "Tried to replace {name} or {resource_name} in location string but name is not set."
            location = location.replace("{name}", name)
            location = location.replace("{resource_name}", name)

        if "{task_name}" in location:
            assert (
                task_name
            ), "Tried to replace {task_name} in location string but task_name is not set."
            location = location.replace("{task_name}", task_name)

        if "{stage}" in location:
            assert (
                stage
            ), "Tried to replace {stage} in location string but stage is not set."
            location = location.replace("{stage}", stage)

        if "{data_version}" in location or "{version}" in location:
            assert (
                data_version
            ), "Tried to replace {data_version} or {version} in location string but data_version is not set."
            location = location.replace("{data_version}", data_version)
            location = location.replace("{version}", data_version)

        if "{code_version}" in location:
            assert (
                code_version
            ), "Tried to replace {code_version} in location string but code_version is not set."
            location = location.replace("{code_version}", code_version)

        return location

    def _parse_name_from_location(self):
        if "http" in str(self.location):
            self._parse_url()
        else:
            self._parse_path()

    def to_pandas(self) -> pd.DataFrame:
        """
        Converts instances of BaseToolModel within BaseTableClass to a pandas DataFrame.

        Returns:
            pandas.DataFrame: DataFrame containing data from instances of BaseToolModel.
        """
        model_dict = self.model_dump()
        model_dict["location"] = self._location
        return pd.DataFrame.from_dict(model_dict, orient="index").T

    def _parse_url(self):
        import urllib

        parsed_source = urllib.parse.urlparse(self.location)
        ls_path_name = parsed_source.path.split("/")
        self.name = ls_path_name[-1] if ls_path_name[-1] else ls_path_name[-2]

    def _parse_path(self):
        self.name = Path(self.location).name

    def append_comment(self, comment: str):
        if self.comment is not None and comment in self.comment:
            logging.warning(
                f"""Comment ignored! 
                The added comment `{comment}` is already in the comment field. No need to add again.
                Please check the comment of the resource {self.name}"""
            )
        elif self.comment is None:
            self.comment = comment
        else:
            self.comment = self.comment + "\n" + comment



import uuid
from datetime import datetime

from pydantic import EmailStr, Field

from dataio.tools import BonsaiBaseModel


class User(BonsaiBaseModel):
    """Pydantic model representing a user."""

    email: EmailStr
    name: str
    is_active: bool = True
    is_staff: bool = False
    is_superuser: bool = False


class DataLicense(BonsaiBaseModel):
    """Pydantic model for Data License."""

    name: str
    description: str | None = None
    url: str
    create_time: datetime = Field(default_factory=datetime.now)
    created_by: User | None = None  # Assuming a ForeignKey-like relationship

    def __str__(self) -> str:
        return self.name


class Version(BonsaiBaseModel):
    version: str
    create_time: datetime = Field(default_factory=datetime.now)
    comments: str | None = None

    def __str__(self) -> str:
        return f"Version {self.version}"


class MetaData(BonsaiBaseModel):
    id: uuid.UUID
    created_by: User
    last_modified: datetime
    license: DataLicense
    version: Version



"""This is an utility module to store data into the group folder of BONSAI project
Created on Jul 15, 2022

@author: Fan Yang

Todo:


"""

import logging
import pickle
from collections import namedtuple
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, List

import frictionless
import pandas as pd
import yaml
from data_io.dataclasses import (
    Algorithm,
    AttributeTable,
    Author,
    DataPackage,
    FactTable,
    License,
    ModifiedBy,
    Source,
)
from frictionless import Package, Resource, describe, extract, validate
from yaml.loader import FullLoader

# self.level = level
# self.logger = logging.getLogger("data_io")
# self.logger.setLevel(level)
# self.logger.addHandler(handler)


def _version_to_path(version_string: str):
    """
    Converts a version string into a pathlib.Path instance.

    Version string needs to be in the format MAJOR.MINOR.PATCH.

    Parameters
    ----------
    version_string : str
        The version string to convert to a Path instance.

    Returns
    -------
    pathlib.Path
        The corresponding pathlib.Path instance.
    """
    versions = version_string.split(".")

    if len(versions) == 2:
        [major, minor] = versions
        return Path(major) / minor

    if len(versions) == 3:
        [major, minor, patch] = versions
        return Path(major) / minor / patch

    raise ValueError(
        f"Malformated version string {version_string}, should be either MAJOR.MINOR.PATCH or MAJOR.MINOR."
    )


def _category_path_to_path(category_path: str):
    """
    Converts a category ID string into a pathlib.Path instance.

    Category ID needs to be in the format SOME/CATEGORY/STACK.

    Parameters
    ----------
    category_path : str
        The category ID string to convert to a Path instance.

    Returns
    -------
    pathlib.Path
        The corresponding pathlib.Path instance.
    """
    error = ValueError(
        f"Invalid category_path {category_path}, should have format SOME/CATEGORY/STACK."
    )

    if "\\" in category_path:
        raise error

    values = category_path.split("/")

    path = Path(values[0])
    for item in values[1:]:
        if not item:
            raise error
        path = path / item
    return path


def write(param_dict, datapackage: DataPackage, perform_checks=True):
    """
    Writes the given DataPackage instance to the file system, performing optional integrity checks.
    The method will create the necessary directories based on the root_path, category_path, and version.
    The DataPackage's tables are written as CSV files, and the metadata is written as a YAML file.
    The output files adhere to the frictionless format, adding some additional metadata.

    Parameters
    ----------
    param_dict : dict
        A dictionary containing the root_path and other parameters. The root_path is used as the base
        directory to store the DataPackage.
    datapackage : DataPackage
        The DataPackage instance to write to the file system.
    perform_checks : bool, optional
        If True (default), performs integrity checks on the DataPackage instance before writing it.
        The checks ensure that the DataPackage has a name, source, modified_by, and target_language.
    """

    datapackage.assert_completeness()

    # TODO: look for non-standard metadata entries and use them to create additional columns
    # TODO:

    if "root_path" in param_dict:
        path = Path(param_dict["root_path"])
        path = (
            path
            / _category_path_to_path(datapackage.category_path)
            / datapackage.name
            / _version_to_path(datapackage.version)
        )

        if not path.exists():
            path.mkdir(parents=True)

        resources = []
        for name, table in datapackage.tables.items():
            file_path = path / f"{name}.csv"
            table.to_csv(file_path)
            resource = frictionless.describe(table.data, name=name)
            resource.data = None
            resource.path = f"{name}.csv"
            resource.scheme = None
            resource.format = "csv"
            resource.mediatype = None
            metadata = resource.metadata_export()
            metadata.update({"tabletype": table.type})

            if isinstance(table, AttributeTable):
                metadata.update(
                    {"attribute_name": table.name, "attribute_category": table.category}
                )

            resource = resource.metadata_import(metadata)
            for index_name in table.data.index.names:
                f = table.fields[index_name]
                if f.foreign_name:
                    foreign_key = {
                        "fields": [index_name],
                        "reference": {
                            "resource": f.resource,
                            "fields": [f.foreign_name],
                        },
                    }
                    if f.external_path:
                        foreign_key["reference"].update(
                            {"external_path": f.external_path}
                        )
                    resource.schema.foreign_keys.append(foreign_key)
            resources.append(resource)

        package = Package(resources=resources)

        source = datapackage.source.__dict__.copy()
        license = source.pop("license")

        authors = [author.__dict__ for author in datapackage.modified_by.authors]

        package = package.metadata_import(
            {
                "name": datapackage.name,
                "category_path": datapackage.category_path,
                "title": datapackage.title,
                "comment": datapackage.comment,
                "version": datapackage.version,
                "last_modified": datapackage.last_modified,
                "modified_by": {
                    "algorithm": datapackage.modified_by.algorithm.__dict__,
                    "authors": authors,
                },
                "license": license.__dict__,
                "source": source,
                "resources": package.resources,
                "target_language": datapackage.target_language,
                "encoding": datapackage.encoding,
            }
        )

        dict = package.to_dict()
        with open(path / f"{datapackage.name}.datapackage.yaml", "w") as file:
            yaml.dump(dict, file)


def read(param_dict: Dict, perform_checks=True):
    """
    Reads a DataPackage from the file system using the provided parameters. By default, it tries to
    load a frictionless data package using a descriptor YAML file. The method reads the metadata
    and the data tables (CSV files) from the specified path.

    Parameters
    ----------
    param_dict : Dict
        A dictionary containing the path and other parameters for reading the DataPackage. The path
        should point to the directory containing the DataPackage's YAML descriptor and CSV files.
    perform_checks : bool, optional
        If True (default), performs integrity checks on the DataPackage instance after reading it.

    Returns
    -------
    DataPackage
        The DataPackage instance read from the file system, with its tables and metadata populated.
    """

    package_name = param_dict["name"]

    if "path" in param_dict:
        fpackage = Package(
            Path(param_dict["path"]) / (package_name + ".datapackage.yaml")
        )
        metadata = fpackage.metadata_export()
        data = fpackage.extract()

        return_package = DataPackage(
            name=package_name,
            category_path=metadata["category_path"],
            version=metadata["version"],
            title=metadata["title"],
            comment=metadata["comment"],
            encoding=metadata["encoding"],
            target_language=metadata["target_language"],
            last_modified=metadata["last_modified"],
        )

        tables = {}
        for name, table in data.items():
            table_metadata = fpackage.get_resource(name).metadata_export()
            tabletype = table_metadata["tabletype"]
            field_metadata = None
            if "foreignKeys" in table_metadata["schema"]:
                field_metadata = table_metadata["schema"]["foreignKeys"]

            if tabletype == FactTable.type:
                tables[name] = FactTable.from_json(table, field_metadata)
            elif tabletype == AttributeTable.type:
                tables[name] = AttributeTable.from_json(
                    table,
                    field_metadata,
                    table_metadata["attribute_name"],
                    table_metadata["attribute_category"],
                )
        return_package.tables = tables

        license_metadata = metadata["license"]
        l = License(
            license_metadata["acronym"],
            license_metadata["full_name"],
            license_metadata["description"],
            license_metadata["link"],
        )

        source_metadata = metadata["source"]
        return_package.source = Source(
            source_metadata["reference"],
            source_metadata["encoding"],
            source_metadata["link"],
            source_metadata["authors"],
            source_metadata["institution"],
            source_metadata["type"],
            l,
            source_metadata["language"],
            source_metadata["link_doc"],
        )

        modified_by_data = metadata["modified_by"]
        authors = []
        for author in modified_by_data["authors"]:
            authors.append(
                Author(author["name"], author["affiliation"], author["e_mail"])
            )
        return_package.modified_by = ModifiedBy(
            authors,
            Algorithm(
                modified_by_data["algorithm"]["name"],
                modified_by_data["algorithm"]["version_number"],
            ),
        )

        return return_package



from typing import Optional

from pydantic import Field

from dataio.schemas.bonsai_api.base_models import DimensionModel


class Level(DimensionModel):
    name: str
    code: str
    description: Optional[str] = None
    comment: Optional[str] = None


class ClassificationNode(DimensionModel):
    code: str
    parent_code: Optional[str] = None
    level: Optional[str] = None
    description: Optional[str] = None
    comment: Optional[str] = None


class Location(ClassificationNode):
    def __str__(self) -> str:
        return self.name


class ActivityType(ClassificationNode):
    # flow_type: str = Field(None, max_length=30)

    def __str__(self) -> str:
        return self.description


class FlowObject(ClassificationNode):
    def __str__(self) -> str:
        return self.description


class Market(ClassificationNode):
    def __str__(self) -> str:
        return self.description


class Unit(DimensionModel):
    scientific_notation: str = Field(..., max_length=50)
    name: Optional[str] = Field(None, max_length=50)
    description: Optional[str] = None
    unit_dimension: str

    def __str__(self) -> str:
        return self.short_name


class UnitConversion(DimensionModel):
    unit: str
    reference_unit: str
    conversion_factor: float


class Calendar(DimensionModel):
    description: str

    def __str__(self) -> str:
        return self.code


class Year(DimensionModel):
    name: str
    calendar: str

    def __str__(self) -> str:
        return self.code


class DataQuality(DimensionModel):
    name: str
    code: str
    description: Optional[str] = None
    comment: Optional[str] = None


class UncertaintyDistribution(DimensionModel):
    name: str
    code: str
    description: Optional[str] = None
    comment: Optional[str] = None


class ChemicalCompound(DimensionModel):
    name: str
    code: str
    description: Optional[str] = None
    comment: Optional[str] = None


class Compartment(DimensionModel):
    name: str
    code: str
    description: Optional[str] = None
    comment: Optional[str] = None


class LCIA(DimensionModel):
    name: str
    code: str
    description: Optional[str] = None
    comment: Optional[str] = None


# Used for fact tables and collection of flags.
class ExternalDimensionTables(DimensionModel):
    description: str
    comment: Optional[str] = None
    urn: Optional[str] = None



import uuid
from datetime import date, datetime
from typing import Optional

##################################### models_fact


class FactRecipe(FactBaseModel):
    new_id: uuid.UUID = Field(default_factory=uuid.uuid4)
    id: int
    flow_reference: str
    region_reference: str
    unit_reference: str
    flow_input: str
    region_inflow: str | None = None
    value_inflow: float | None = None
    unit_inflow: str | None = None
    value_emission: float
    unit_emission: str
    metrics: str

    class Config:
        from_attributes = True


class UnitConverter(BaseModel):
    unit_from_update: str | None = (
        None  # Assuming this is a foreign key relationship in Django
    )
    unit_to_update: str | None = (
        None  # Assuming this is a foreign key relationship in Django
    )
    unit_from: str
    unit_to: str
    multiplier: float
    unit_type: str

    def __str__(self) -> str:
        return f"{self.unit_from} - {self.unit_to}"

    class Config:
        from_attributes = True


#################################### models_dim


class DimProductDjango(ModelDimDjango):
    code = models.CharField(max_length=20, null=False, primary_key=True)
    name = models.CharField(max_length=200, null=True, unique=True)
    parent_name = models.ForeignKey(
        "self",
        to_field="name",
        null=True,
        related_name="children",
        on_delete=models.SET_NULL,
    )
    description = models.TextField(null=True, unique=True)

    def __str__(self) -> str:
        return f"{self.description}"

    class Meta:
        db_table = "dim_product"
        verbose_name = "Product Classification"


############################################ models_corr


class CorrespondenceDjango(models.Model):
    created_by = models.ForeignKey(
        to=UserDjango,
        to_field="email",
        related_name="%(class)s",
        on_delete=models.SET_NULL,
        null=True,
    )
    create_time = models.DateField(null=False, default=date.today)

    class Meta:
        abstract = True


class ProductCorrespondenceDjango(CorrespondenceDjango):
    external_name = models.CharField(null=False, max_length=200)
    external_description = models.TextField(null=True)
    base_name = models.ForeignKey(
        to=DimProduct, to_field="name", on_delete=models.SET_NULL, null=True
    )

    def __str__(self) -> str:
        return f"{self.external_name} - {self.base_name}"

    class Meta:
        db_table = "corr_product"
        verbose_name = "Product Correspondence"


############################# datahub


class DatasetMeta(BaseModel):
    name: str = Field(..., max_length=255)
    acronym: str = Field(..., max_length=10)
    version_date: date
    version_semantic: Optional[str] = Field(None, max_length=10)
    is_latest: bool = True
    create_time: datetime = Field(default_factory=datetime.now)
    created_by: Optional[User] = None
    comment: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.name} - {self.acronym}"


class DatasetMetaDjango(models.Model):
    name = models.CharField(max_length=255, null=False)
    acronym = models.CharField(max_length=10, null=False)
    version_date = models.DateField(null=False)
    version_semantic = models.CharField(max_length=10, null=True)
    is_latest = models.BooleanField(default=True, null=False)
    create_time = models.DateTimeField(null=False, auto_now_add=True)
    created_by = models.ForeignKey(
        to=UserDjango,
        to_field="email",
        related_name="dataset",
        on_delete=models.SET_NULL,
        null=True,
    )
    comment = models.TextField(null=True)

    class Meta:
        db_table = "dataset_meta"
        verbose_name = "Meta Data"
        verbose_name_plural = "Meta Data"


################# footprint_analyzer


class CountryFootprintDjango(FactBaseModel):
    act_code = models.ForeignKey(
        to=DimActivityDjango, to_field="code", null=False, on_delete=models.CASCADE
    )
    region_code = models.ForeignKey(
        to=DimRegionDjango, to_field="code", null=False, on_delete=models.CASCADE
    )
    value = models.FloatField(null=False)
    unit_emission = models.ForeignKey(
        to=DimUnitDjango,
        to_field="short_name",
        null=False,
        default="tonnes CO2eq",
        on_delete=models.CASCADE,
    )

    def __str__(self) -> str:
        return f"{self.act_code}-{self.region_code}"

    class Meta:
        db_table = "fact_country_footprint"
        verbose_name = "Country Footprint"


class CountryRecipeDjango(FactBaseModelDjango):
    product_code = models.ForeignKey(
        to=DimProductDjango,
        to_field="code",
        null=True,
        on_delete=models.SET_NULL,
    )
    unit_reference = models.ForeignKey(
        to=DimUnitDjango,
        to_field="short_name",
        null=True,
        on_delete=models.SET_NULL,
        related_name="unit_reference",
    )
    act_code = models.ForeignKey(
        to=DimActivityDjango, to_field="code", null=False, on_delete=models.CASCADE
    )
    region_code = models.ForeignKey(
        to=DimRegionDjango, to_field="code", null=False, on_delete=models.CASCADE
    )
    value = models.FloatField(null=False)
    unit_emission = models.ForeignKey(
        to=DimUnitDjango,
        to_field="short_name",
        null=False,
        default="tonnes CO2eq",
        on_delete=models.CASCADE,
        related_name="unit_emission",
    )

    class Meta:
        db_table = "fact_ctry_recipe"
        verbose_name = "Country Footprint Recipe"



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 16:05:31 2022

@author: ReMarkt
"""
import copy
import json
import logging
import os
from dataclasses import fields
from pathlib import Path

import yaml
from dotenv import load_dotenv

from dataio.resources import ResourceRepository
from dataio.utils.path_manager import PathBuilder

logger = logging.getLogger("config")


class Config:

    @staticmethod
    def get_airflow_defaults(filename: str):
        """
        Load config from airflow or from src/config/airflow_attributes.config.json
        Args:
            filename (str): JSON config file located in the folder 'src/config'
        Returns: None
        """
        config_path = Path(__file__).parent.parent / "config" / filename

        logger.debug("path= ", str(config_path))
        try:
            with open(config_path, "r") as file:
                return json.load(file)
        except FileNotFoundError:
            raise Exception(
                f"Config file '{filename}' not found, please create it in the folder src/config"
            )

    def __init__(
        self,
        current_task_name=None,
        custom_resource_path=None,
        log_level=logging.WARNING,
        log_file=None,
        run_by_user=None,
        **kwargs,
    ):
        """
        Config class for managing configurations, including logging.

        :param current_task_name: Name of the current task (used for logging).
        :param custom_resource_path: Optional path for custom resources.
        :param log_level: Logging level (default: WARNING).
        :param log_file: Optional file path to log output.
        :param kwargs: Additional configuration parameters.
        """

        self.current_task_name = current_task_name
        self.custom_resource_path = custom_resource_path
        self.run_by_user = run_by_user 

        self._load_config("data_attributes.config.yaml")
        self._load_airflow_config("airflow_attributes.config.json")

        # Set up logging
        self.log_handler = logging.StreamHandler()
        self.log_level = log_level
        self.logger = self._setup_logger(log_file)

        # Assign additional configuration attributes
        for key, value in kwargs.items():
            setattr(self, key, value)
            if key == "config_path":
                self._load_config(value, True)

    def _setup_logger(self, log_file):
        """
        Sets up the logger with a console handler and optional file handler.
        """
        logger = logging.getLogger(self.current_task_name)
        logger.setLevel(self.log_level)

        # Prevent duplicate log handlers
        if not logger.handlers:
            # Console handler
            console_handler = logging.StreamHandler()
            console_handler.setLevel(self.log_level)

            # Formatter
            formatter = logging.Formatter(
                "[%(asctime)s] [%(levelname)s] [%(name)s]: %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            console_handler.setFormatter(formatter)
            logger.addHandler(console_handler)

            # File handler (if provided)
            if log_file:
                file_handler = logging.FileHandler(log_file)
                file_handler.setLevel(self.log_level)
                file_handler.setFormatter(formatter)
                logger.addHandler(file_handler)

        return logger

    def log(self, level, message):
        """
        Logs a message with the specified logging level.
        """
        if self.logger:
            self.logger.log(level, message)

    def load_env(self) -> dict:
        load_dotenv()
        env_dict = {}

        for key in os.environ:
            env_dict[key] = os.environ[key]
        return env_dict

    @property
    def bonsai_home(self):
        """environment variable to define the home directory for dataio"""
        env_dict = self.load_env()
        assert env_dict[
            "BONSAI_HOME"
        ], "Please set up environmental variable for 'BONSAI_HOME'"
        return Path(env_dict.get("BONSAI_HOME", str(Path.home())))

    @property
    def dataio_root(self):
        """environment variable to define the home directory for hybrid_sut"""
        env_dict = self.load_env()
        assert env_dict[
            "DATAIO_ROOT"
        ], "Please set up environmental variable for 'DATAIO_ROOT'"
        return Path(env_dict.get("DATAIO_ROOT", str(Path.home())))

    @property
    def path_repository(self) -> PathBuilder:

        from dataio.utils.accounts import AccountRepository

        print(f"Get version from {self.version_source}")
        return PathBuilder(
            Path(self.bonsai_home),
            version_source=self.version_source,
            account_repository=AccountRepository(
                self.bonsai_home / "_bonsai" / "accounts.json"
            ),
        )

    @property
    def version_source(self):
        vdate = self.date.replace("-", "")

        path_version = (
            self.bonsai_home / "_bonsai" / "versions" / f"versions_{vdate}.txt"
        )
        if not path_version.exists():
            path_version = self.bonsai_home / "versions" / f"versions_{vdate}.txt"

        # TODO: replace this path with Version class
        return path_version

    @version_source.setter
    def version_source(self, path: Path) -> None:
        self.version_source = path

    @property
    def schemas(self):
        from dataio.schemas import bonsai_api

        return bonsai_api

    @property
    def schema_enums(self):
        from dataio.utils import schema_enums

        return schema_enums

    @property
    def connector_repository(self):
        from dataio.utils.connectors import connector_repository

        return connector_repository

    def list_parameters(self):
        """List all dataclass field names."""

        return [field.name for field in fields(self)]

    @property
    def resource_repository(self) -> ResourceRepository:
        from dataio.resources import ResourceRepository

        db_path = (
            self.custom_resource_path if self.custom_resource_path else self.dataio_root
        )

        return ResourceRepository(db_path=db_path)

    @property
    def sut_resource_repository(self):
        from dataio.utils.hsut.resources_hsut import (
            CSVResourceRepository as SutCSVResourceRepository,
        )

        db_path = (
            self.path_repository.exiobase4
            if self.path_repository.exiobase4
            else self.bonsai_home
        )

        return SutCSVResourceRepository(db_path)

    def _load_config(self, filename: str, isLocalfile: bool = False) -> None:
        """Load config file in the class Config's attributes
        Args:
            filename (str): YAML config file located in the folder 'src/config'
        Returns: None
        """
        config_path = (
            Path(filename)
            if isLocalfile
            else Path(__file__).parent.parent / "config" / filename
        )

        logger.debug("path= ", {config_path})
        try:
            with open(config_path, "r") as file:
                config = yaml.safe_load(file)
                for key, value in config.items():
                    self.__setattr__(key, value)
            return None
        except FileNotFoundError:
            raise Exception(
                f"Config file '{filename}' not found, please create it, or one in the folder src/config"
            )

    def _load_airflow_config(self, filename: str):
        """
        Load config from airflow or from src/config/airflow_attributes.config.json
        Args:
            filename (str): JSON config file located in the folder 'src/config'
        Returns: None
        """

        airflow_defaults = Config.get_airflow_defaults(filename)
        for key, value in airflow_defaults.items():
            self.__setattr__(key, value)

    def copy(self):
        """
        Creates a deep copy of the current Config instance.
        """
        return copy.deepcopy(self)



from logging import getLogger
from pathlib import Path

import pytest
import yaml

from dataio.config import Config

logger = getLogger("unit_test")


def test_load_config_from_yaml() -> None:
    config = Config("current_task_name")
    assert config.fao_cfile == "20230404", "error in global file"
    assert config.pasture == [
        "Land under perm. meadows and pastures",
        "Land under temp. meadows and pastures",
        "Permanent meadows and pastures",
        "Temporary meadows and pastures",
    ], "error in data file"
    assert config.retailer_sec["Retailer"] == [
        "A_TDRT",
        "A_TDWH",
    ], "error in data file subdivision"
    return None


def test_load_config_from_localfile() -> None:
    """Test import of a config file from local computer"""
    config_file_path = Path(__file__).parent / "test" / "global.config.test.yaml"
    task_name = "current_task_name"
    fao_cfile_value = "aaaaaaa"

    config = Config(task_name, config_path=config_file_path)
    assert (
        config.fao_cfile == fao_cfile_value
    ), "error in global file"  # add the key 'fao_cfile' and the value you want to test
    return None



import ast
import json
import os
import types
from typing import Any, Dict, List, Optional, Type, Union, get_args, get_origin

import pandas as pd
from pydantic import BaseModel, TypeAdapter


class BonsaiBaseModel(BaseModel):

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs):
        super().__pydantic_init_subclass__(**kwargs)
        # Create a new type for names dynamically
        fields = list(cls.model_fields.keys())
        fields = fields + list(cls.model_computed_fields.keys())
        cls.names = type("Names", (), {field: field for field in fields})

    def to_pandas(self) -> pd.DataFrame:
        """
        Converts instances of BaseToolModel within BaseTableClass to a pandas DataFrame.

        Returns:
            pandas.DataFrame: DataFrame containing data from instances of BaseToolModel.
        """
        return pd.DataFrame.from_dict(self.model_dump(), orient="index").T

    @classmethod
    def to_dataclass(cls: Type["BaseModel"], input_data) -> "BonsaiTableModel":
        adapter = TypeAdapter(list[cls])  # Compile once for list of cls

        if isinstance(input_data, pd.DataFrame):
            data_dicts = input_data.to_dict(orient="records")
        elif isinstance(input_data, dict):
            data_dicts = input_data.get("data", [])
        elif isinstance(input_data, pd.MultiIndex):
            data_dicts = [dict(zip(input_data.names, idx)) for idx in input_data]
        else:
            raise ValueError(
                "Invalid input. Must be a pandas DataFrame, a JSON object, or a MultiIndex."
            )

        validated_data = adapter.validate_python(data_dicts)
        return BonsaiTableModel(data=validated_data)

    @classmethod
    def transform_dataframe(
        cls: Type["BonsaiBaseModel"],
        df: pd.DataFrame,
        column_mapping: Optional[Dict[str, str]] = None,
    ) -> pd.DataFrame:
        """
        Transform a DataFrame into the BonsaiTable format by renaming columns and dropping others.
        Ensures all non-optional columns are present, and keeps optional columns if they are in the DataFrame.
        The column_mapping is optional. If not provided, only columns matching the schema fields are kept.

        Args:
            df (pd.DataFrame): The input DataFrame to transform.
            column_mapping (Optional[Dict[str, str]]): A dictionary mapping input DataFrame column names to schema field names.
                If None, only columns matching the schema fields will be kept.

        Returns:
            pd.DataFrame: A DataFrame with columns renamed (if mapping is provided) and unnecessary columns dropped.

        Raises:
            ValueError: If any required columns are missing from the DataFrame.
        """
        # Define non-optional and optional columns
        non_optional_fields = {
            name
            for name, info in cls.model_fields.items()
            if not (
                get_origin(info.annotation) is Union
                and type(None) in get_args(info.annotation)
            )
        }
        optional_fields = {
            name
            for name, info in cls.model_fields.items()
            if get_origin(info.annotation) is Union
            and type(None) in get_args(info.annotation)
        }

        if column_mapping:
            # Create a mapping dictionary to rename columns
            rename_mapping = {
                input_col: schema_field
                for input_col, schema_field in column_mapping.items()
            }

            # Rename the columns
            df = df.rename(columns=rename_mapping)

        # Collect columns that are present in the DataFrame
        columns_present = set(df.columns)

        # Ensure all non-optional columns are present
        missing_required_cols = non_optional_fields - columns_present
        if missing_required_cols:
            raise ValueError(
                f"Missing required columns: {', '.join(missing_required_cols)}"
            )

        # Include optional columns if they are present
        columns_to_keep = non_optional_fields.union(optional_fields).intersection(
            df.columns
        )
        columns_to_keep = list(columns_to_keep)
        df_transformed = df[columns_to_keep]

        # Return the transformed DataFrame
        return df_transformed

    @classmethod
    def get_csv_field_dtypes(cls: Type["BaseModel"]) -> Dict[str, Any]:
        """
        Return a dictionary with field names and their corresponding types.
        Since csv files can only contain str, float and int, all types that
        are not int and float will be changed to str
        """

        def convert_dtype(dtype_str):
            if dtype_str in ["str", "float", "int", "bool"]:
                return dtype_str
            else:
                return "str"

        type_dict = {}
        for field_name, field_info in cls.model_fields.items():
            field_type = field_info.annotation
            origin = get_origin(field_type)
            if origin is Union or isinstance(field_type, types.UnionType):
                # Check for Optional types (Union[X, None])
                args = get_args(field_type)
                if len(args) == 2 and type(None) in args:
                    # Get the non-None type
                    primary_type = next(arg for arg in args if arg is not type(None))
                    type_name = convert_dtype(primary_type.__name__)
                else:
                    # Otherwise, list all types
                    type_name = ", ".join(convert_dtype(arg.__name__) for arg in args)
            else:
                type_name = convert_dtype(field_type.__name__)
            type_dict[field_name] = type_name

        for field_name, field_info in cls.model_computed_fields.items():
            field_type = field_info.return_type
            type_dict[field_name] = field_type.__name__

        return type_dict

    @classmethod
    def get_empty_dataframe(cls: Type["BonsaiBaseModel"]):
        """
        Returns an empty pandas DataFrame with columns corresponding to the fields of the data class.

        Returns:
            pandas.DataFrame: An empty DataFrame with columns corresponding to the fields of the data class.
        """
        columns = list(cls.model_fields.keys())
        columns.extend(cls.model_computed_fields.keys())

        df = pd.DataFrame(columns=columns)
        return df.astype(cls.get_csv_field_dtypes())

    @classmethod
    def get_classification(cls) -> Dict[str, str]:
        """
        Retrieves the classification dictionary, hidden from serialization.
        """
        return cls._classification

    @classmethod
    def get_api_endpoint(cls) -> Dict[str, str]:
        """
        Retrieves the api endpoint dictionary, hidden from serialization.
        """
        return cls._endpoints


class BonsaiTableModel(BaseModel):
    data: list[BonsaiBaseModel]
    # schema_name: str

    def to_json(self):
        """
        Convert the object to a JSON string representation.

        Returns:
            str: A JSON string representing the object with data information.
        """
        return json.dumps(
            {
                "data": [item.model_dump() for item in self.data],
            },
            indent=4,
        )

    def to_pandas(self) -> pd.DataFrame:
        """
        Converts instances of BaseToolModel within BaseTableClass to a pandas DataFrame.

        Returns:
            pandas.DataFrame: DataFrame containing data from instances of BaseToolModel.
        """
        # Ensure consistent order of columns
        data_dict = {key: [] for key in self.data[0].__dict__.keys()}

        for item in self.data:
            for key, value in item.model_dump().items():
                data_dict[key].append(value)

        return pd.DataFrame(data_dict)


def get_dataclasses(directory: str = "src/dataio/schemas/bonsai_api") -> List[str]:
    """
    Retrieve a list of Pydantic dataclass names that inherit from BaseToolModel from Python files in the specified directory.

    Args:
        directory (str): The directory path where Python files containing Pydantic dataclasses are located.
            Defaults to "src/dataio/schemas/bonsai_api".

    Returns:
        List[str]: A list of fully qualified names (including module names) of Pydantic dataclasses that inherit from BaseToolModel.
    """
    dataclasses = []
    found_classes = set()

    # Iterate through each file in the directory
    for filename in os.listdir(directory):
        if filename.endswith(".py"):
            file_path = os.path.join(directory, filename)
            with open(file_path, "r") as file:
                try:
                    tree = ast.parse(file.read(), filename=file_path)
                except SyntaxError:
                    print(f"Error parsing file: {file_path}")
                    continue

            # Find class definitions in the abstract syntax tree
            for node in ast.walk(tree):
                if isinstance(node, ast.ClassDef):
                    class_name = node.name
                    # qualified_name = f"{filename[:-3]}.{class_name}"
                    qualified_name = f"{class_name}"
                    # Check if it's a Pydantic dataclass and inherits from BaseToolModel
                    if any(
                        isinstance(base, ast.Name)
                        and base.id == "BonsaiBaseModel"
                        or base.id in dataclasses
                        for base in node.bases
                    ):
                        if qualified_name not in dataclasses:
                            dataclasses.append(qualified_name)
                            found_classes.add(qualified_name)

    return dataclasses


def print_data_classes():
    """
    Print out all the available data classes in the directory src/dataio/schemas/bonsai_api
    """
    list_dataclasses = get_dataclasses()
    print(list_dataclasses)
    return



"""Tutorial about executing data transformation task."""

from dataio import load, plot, save, validate

print("loading metadata of task")
task = load(full_path="task.dataio.yaml", include_tables=False)

print("loading dependencies")
lookup_path = task.__metadata__["datapackages"][0]["path"]
lookup_name = task.__metadata__["datapackages"][0]["name"]
lookup = load(full_path=f"{lookup_path}/{lookup_name}.dataio.yaml", include_tables=True)
del lookup_name, lookup_path

print("populating tables (example)")
index = lookup.tree_main.loc[lookup.tree_main["dim_level"] == 2].index
for pos, val in enumerate(list(index)):
    print(pos, val)
    task.dim_dimension.loc[val] = {"position": pos}
    task.fact_fact.loc[str(pos)] = {"dim_dimension": val, "value": float(pos)}


print("saving")
save(datapackage=task, increment="minor", overwrite=True, create_path=True)

print("validating")
validate(full_path="task/task.dataio.yaml", overwrite=True, log_name="validate.log")

print("plotting")
plot(full_path="task/task.dataio.yaml", overwrite=True)



from logging import getLogger
from shutil import copytree

import yaml

from dataio.config import Config

from .set_logger import set_logger

logger = getLogger("root")


def arg_parser(package: object, main: object, argv: list):
    """Argument parser of main task executable.

    Parameters
    ----------
    package : obj
    main: obj
    argv: list
    """

    set_logger()
    logger.info(f"Executing Python package '{package.__name__}'")

    # validate package metadata
    for attr in ["STAGE", "TASKS"]:
        if attr not in dir(package):
            logger.error(f"Attribute '{attr}' is missing from package")
            raise

    help_str = (
        f"usage: python -m {package.__name__} [-h] [-r YAML] [-e PATH]"
        "\n"
        "\noptions:"
        "\n  -h, --help             show this help message and exit"
        "\n  -r , --run RUN         path to configuration YAML file"
        "\n  -e, --export EXPORT    path to store exported data"
        "\n                           (returns error if path exists)"
        "\n  -o, --overwrite EXPORT path to store exported data"
        "\n                           (overwrites if path exists)"
        "\n"
    )

    if len(argv) == 1:
        logger.warning("Not enough arguments provided:\n" f"{help_str}")
    else:
        # run
        if argv[1] in ["-r", "--run"]:
            if len(argv) == 2:
                logger.warning(
                    "No configuration file __name__ provided, "
                    "assumed default = 'config.yaml'"
                )
                argv.append("config.yaml")

            try:
                with open(argv[2], "r") as stream:
                    config = Config(**yaml.safe_load(stream))
                logger.info(f"Read configuration file '{argv[2]}'")
            except FileNotFoundError:
                logger.error(
                    "Error loading configuration yaml "
                    f"'{argv[2]}'. Maybe a path or file format "
                    "issue."
                )
                raise
            main(config=config)
        # export
        elif argv[1] in ["-e", "--export"]:
            if len(argv) == 2:
                argv.append(".")
            logger.info(
                f"Exporting data from package "
                f"{package.__name__} to path '{argv[2]}'."
            )
            copytree(package.EXPORT_PATH, argv[2], dirs_exist_ok=False)
            logger.info("Finished exporting data.")
        # overwrite
        elif argv[1] in ["-o", "--overwrite"]:
            if len(argv) == 2:
                argv.append(".")
            logger.info(
                f"Exporting data from package "
                f"{package.__name__} to path '{argv[2]}'."
            )
            copytree(package.EXPORT_PATH, argv[2], dirs_exist_ok=True)
            logger.info("Finished exporting data.")
        elif argv[1] in ["-h", "--help"]:
            logger.info(help_str)
        else:
            logger.info("Invalid arguments provided.")
            logger.info(help_str)



import datetime
from typing import ClassVar, Dict, Optional, Tuple

from pydantic import Field, PrivateAttr

import dataio.schemas.bonsai_api.PPF_fact_schemas_uncertainty as PPF_fact_schemas_uncertainty
from dataio.schemas.bonsai_api.base_models import FactBaseModel_uncertainty


class PRODCOMProductionVolume(
    PPF_fact_schemas_uncertainty.ProductionVolumes_uncertainty
):
    indicator: str

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("geonumeric", "location"),
        "product": ("prodcom_total_2_0", "flowobject"),
    }  # Total production (ds-056121)

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"


class PRODCOMSoldProductionVolume(
    PPF_fact_schemas_uncertainty.ProductionVolumes_uncertainty
):
    indicator: str

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("geonumeric", "location"),
        "product": ("prodcom_sold_2_0", "flowobject"),
    }  # Sold production (ds-056120)

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"


class IndustrialCommodityStatistic(
    PPF_fact_schemas_uncertainty.ProductionVolumes_uncertainty
):

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("regex", "location"),
        "product": ("undata_ics", "flowobject"),
    }

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"


class baseExternalSchemas(FactBaseModel_uncertainty):
    # Created with Sanders. Might Not be the base
    location: str
    time: int
    unit: str
    value: float
    comment: Optional[str] = None
    flag: Optional[str] = None


class ExternalMonetarySUT(baseExternalSchemas):
    table_type: str  # Supply or use table
    product_code: str
    product_name: str
    activity_code: str
    activity_name: str
    price_type: str = Field(  # Current price or previous year price.
        default="current prices"
    )
    consumer_price: bool = Field(default=False)  # Consumer price vs production price
    money_unit: Optional[str] = None  # Millions, billions etc.
    diagonal: Optional[bool] = None


# For annual data tables that start on a day other than January 1st. E.g. the fiscal year of India.
class BrokenYearMonetarySUT(ExternalMonetarySUT):
    time: datetime.date  # Start date of fiscal year


class EuropeanMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("cpa_2_1_lvl2", "flowobject"),
        "activity_code": ("nace_rev2", "activitytype"),
    }


class OldEuropeanMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("cpa_2008", "flowobject"),
        "activity_code": ("nace_rev2", "activitytype"),
    }


class OlderEuropeanMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("cpa_2008", "flowobject"),
        "activity_code": ("nace_rev1_1", "activitytype"),
    }


class InternationalMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("cpc_2_1", "flowobject"),
        "activity_code": ("isic_rev4", "activitytype"),
    }


class NACEMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("nace_rev2", "flowobject"),
        "activity_code": ("nace_rev2", "activitytype"),
    }


class OECDMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("cpa_2008", "flowobject"),
        "activity_code": ("isic_rev4", "activitytype"),
    }


class AfricanMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("africa_flow", "flowobject"),
        "activity_code": ("isic_africa", "activitytype"),
    }


class AustralianMonetarySUT(ExternalMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("suic", "flowobject"),
        "activity_code": ("anzsic_2006", "activitytype"),
    }


class EgyptianMonetarySUT(BrokenYearMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("cpc_1_1", "flowobject"),
        "activity_code": ("isic_rev4", "activitytype"),
    }


class IndianMonetarySUT(BrokenYearMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("india_sut", "flowobject"),
        "activity_code": ("india_sut", "activitytype"),
    }


class JapanMonetarySUT(BrokenYearMonetarySUT):
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product_code": ("japan_sut", "flowobject"),
        "activity": ("japan_sut", "activitytype"),
    }


class FAOstat(FactBaseModel_uncertainty):
    product_name: str
    product: str
    fao_element: str
    location: str
    time: int
    value: float
    unit: str
    flag: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("faostat", "location"),
        "product": ("faostat", "flowobject"),
    }


class FAOtrade(FactBaseModel_uncertainty):
    product_name: str
    product: str
    export_location: str  # Exporter
    import_location: str  # Importer
    time: int
    value: float
    unit: str
    flag: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("faostat", "location"),
        "partner": ("fao_area", "location"),
        "product": ("fao_items", "flowobject"),
    }


class UNdataEnergyBalance(FactBaseModel_uncertainty):
    activity: str
    product: str
    location: str
    time: int
    value: float
    unit: str
    comment: Optional[str] = None
    flag: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("iso_3166_1_numeric", "location"),
        "product": ("undata_energy_stats", "flowobject"),  # or prodcom_sold_2_0
    }

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}"


class UNdataEnergyStatistic(FactBaseModel_uncertainty):
    activity: str
    product: str
    location: str
    time: int
    value: float
    unit: str
    comment: Optional[str] = None
    flag: Optional[str] = None
    conversion_factor: Optional[float] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("iso_3166_1_numeric", "location"),
        "product": ("undata_energy_stats", "flowobject"),
        "activity": ("undata_energy_stats", "activitytype"),
    }

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}"


class BACITrade(FactBaseModel_uncertainty):
    time: int
    product: str
    export_location: str
    import_location: str
    value: float
    unit: str
    flag: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product": ("hs1992", "flowobject"),
        "export_location": ("iso_3166_1_alpha2", "location"),
        "import_location": ("iso_3166_1_alpha2", "location"),
    }


class USGSProductionVolume(PPF_fact_schemas_uncertainty.ProductionVolumes_uncertainty):

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("usgs", "location"),
        "product": ("usgs", "flowobject"),
    }

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"


class StatCanChemProductionVolume(
    PPF_fact_schemas_uncertainty.ProductionVolumes_uncertainty
):

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product": ("statcan_chemicals", "flowobject"),
        "location": ("iso_3166_1_alpha2", "location"),
    }

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"


class ComexTrade(FactBaseModel_uncertainty):
    time: int
    product: str
    activity: str
    export_location: str
    import_location: str
    value: float
    unit: str
    transport_mode: Optional[str] = None
    country_active_transport_mode: Optional[str] = None
    container: Optional[str] = None
    flag: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product": ("comext", "flowobject"),
        "activity": ("comext", "activitytype"),
        "export_location": ("iso_3166_1_alpha2", "location"),
        "import_location": ("iso_3166_1_alpha2", "location"),
    }


class UNdataWDI(FactBaseModel_uncertainty):
    location: str
    time: int
    value: float
    unit: str
    flag: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("iso_3166_1_alpha3", "location"),
    }


class ProdcomTrade(FactBaseModel_uncertainty):
    time: int
    product: int
    export_location: str
    import_location: str
    value: float
    unit: str
    flag: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "product": ("prodcom_sold_2_0", "flowobject"),
        "export_location": ("geonumeric", "location"),
        "import_location": ("geonumeric", "location"),
    }


class UnfcccProductionVolume(
    PPF_fact_schemas_uncertainty.ProductionVolumes_uncertainty
):

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("iso_3166_1_alpha3", "location"),
        "product": (
            "unfccc_prodvol",
            "flowobject",
        ),  # I'd  add in the correspondence next sprint
    }

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"


class ADBMonetaryIOT(baseExternalSchemas):
    supplier_name: str  # Supplying activity
    supplier_code: str
    user_name: str  # Using activity
    user_code: str
    price_type: str = Field(  # Current price or previous year price.
        default="current prices"
    )
    money_unit: Optional[str] = None  # Millions, billions etc.
    diagonal: Optional[bool] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "supplier_code": ("adb_supplier", "flow"),
        "user_code": ("adb_user", "flow"),
        "location": ("adb", "location"),
    }



from typing import ClassVar, Dict, Optional, Tuple

from pydantic import Field

import dataio.schemas.bonsai_api.facts as schemas
from dataio.schemas.bonsai_api.base_models import FactBaseModel_samples
from dataio.tools import BonsaiTableModel


class Use_samples(FactBaseModel_samples):
    location: str
    product: str
    activity: str
    unit: str
    value: float
    associated_product: Optional[str] = None
    flag: Optional[str] = (
        None  # TODO flag rework. Can be uncertainty, can be other. Different meanings across data sources.
    )
    time: int
    product_origin: Optional[str] = None  # Where the used product comes from.
    product_type: str = Field(
        default="use"
    )  # set automatically based on what data class is used
    account_type: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.product_type}-{self.activity}-{self.time}-{self.value}-{self.unit}"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "use-samples/": "Endpoint for use samples for both list and detail view.",
    }


class Supply_samples(FactBaseModel_samples):
    location: str
    product: str
    activity: str
    unit: str
    value: float
    product_destination: Optional[str] = None
    associated_product: Optional[str] = None
    flag: Optional[str] = (
        None  # TODO flag rework. Can be uncertainty, can be other. Different meanings across data sources.
    )
    time: int
    product_type: str = Field(
        default="supply"
    )  # set automatically based on what data class is used. This can also be joint or combined product, but maybe needs to be a different attribute?
    account_type: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.product_type}-{self.activity}-{self.time}-{self.value}-{self.unit}"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "supply-samples/": "Endpoint for supply samples for both list and detail view.",
    }


class Imports_samples(FactBaseModel_samples):
    location: str
    product: str
    product_origin: str  # Where the product comes from
    unit: str
    value: float
    time: int
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "import-samples/": "Endpoint for import samples for both list and detail view.",
    }


class Valuation_samples(FactBaseModel_samples):
    location: str
    product: str
    valuation: str
    unit: str
    value: float
    time: int
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "valuation-samples/": "Endpoint for valuation samples for both list and detail view.",
    }


class FinalUses_samples(FactBaseModel_samples):
    location: str
    product: str
    final_user: str  # Final use ctivity that uses the product
    unit: str
    value: float
    time: int
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "final-use-samples/": "Endpoint for final-use samples for both list and detail view.",
    }


class SUTAdjustments_samples(FactBaseModel_samples):
    location: str
    adjustment: str
    product: Optional[str] = None
    product_origin: Optional[str] = None  # Where the product comes from
    final_user: Optional[str] = None  # Where the product is used
    unit: str
    value: float
    time: int

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "sut-adjustment-samples/": "Endpoint for sut-adjustment samples for both list and detail view.",
    }


class OutputTotals_samples(FactBaseModel_samples):
    location: str
    activity: str
    output_compartment: str  # Where the outputs are used
    unit: str
    value: float
    time: int
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "output-total-samples/": "Endpoint for output-total samples for both list and detail view.",
    }


class ValueAdded_samples(FactBaseModel_samples):
    location: str
    activity: str
    value_added_component: str  # Component of value added
    unit: str
    value: float
    time: int

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "value-added-samples/": "Endpoint for value-added samples for both list and detail view.",
    }


class SocialSatellite_samples(FactBaseModel_samples):
    location: str
    activity: str
    social_flow: str  # Type of social flow
    unit: str
    value: float
    time: int

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "social-satellite-samples/": "Endpoint for social-satellite samples for both list and detail view.",
    }


class ProductionVolumes_samples(FactBaseModel_samples):
    location: str
    product: str
    activity: Optional[str] = None
    unit: str
    value: float
    flag: Optional[str] = None  # TODO flag rework
    time: int
    inventory_time: Optional[int] = None
    source: Optional[str] = None
    comment: Optional[str] = None
    price_type: Optional[str] = None
    account_type: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.value}-{self.unit}"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "production-volume-samples/": "Endpoint for production-volume samples for both list and detail view.",
    }


class Emissions_samples(FactBaseModel_samples):
    time: int
    year_emission: Optional[int] = (
        None  # TODO Rework into how we want to handle delayed emissions
    )
    location: str
    activity: str
    activity_unit: str
    emission_substance: str
    compartment: str  # location of emission, such as "Emission to air"
    product: str
    product_unit: str
    value: float
    unit: str
    flag: Optional[str] = None

    elementary_type: str = Field(default="emission")

    def __str__(self) -> str:
        return f"{self.location}-{self.emission_substance}-{self.activity}-{self.activity_unit}-{self.time}-{self.value}-{self.unit}"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "emission-samples/": "Endpoint for emission samples for both list and detail view.",
    }


class TransferCoefficient_samples(FactBaseModel_samples):  # Similar to use
    location: Optional[str] = None
    output: str
    input_product: str
    activity: str
    coefficient_value: float
    unit: str
    flag: Optional[str] = (
        None  # TODO flag rework. Can be uncertainty, can be other. Different meanings across data sources.
    )
    time: Optional[int] = None
    transfer_type: str  # Should be one of these three value: "product", "emission" or "waste" TODO Validator

    def __str__(self) -> str:
        return f"{self.location}-{self.product}-{self.activity}-{self.time}-{self.coefficient_value}"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "transfer-coefficient-samples/": "Endpoint for transfer-coefficient samples for both list and detail view.",
    }


class Resource_samples(Emissions_samples):
    def __init__(self, **data):
        super().__init__(**data)
        self.elementary_type = "resource"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "resource-samples/": "Endpoint for resource samples for both list and detail view.",
    }


class PackagingData_samples(Supply_samples):
    def __init__(self, **data):
        super().__init__(**data)
        self.product_type = "packaging_data"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "packaging-data-samples/": "Endpoint for packaging-data samples for both list and detail view.",
    }


class WasteUse_samples(Use_samples):
    waste_fraction: bool

    def __init__(self, **data):
        super().__init__(**data)
        self.product_type = "waste_use"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "waste-use-samples/": "Endpoint for waste-use samples for both list and detail view.",
    }


class WasteSupply_samples(Supply_samples):
    waste_fraction: bool

    def __init__(self, **data):
        super().__init__(**data)
        self.product_type = "waste_supply"

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product_code": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "waste-supply-samples/": "Endpoint for waste-supply samples for both list and detail view.",
    }


class PropertyOfProducts_samples(FactBaseModel_samples):
    location: Optional[str] = None
    product: str
    value: float
    activity: Optional[str] = None
    unit: str
    description: Optional[str] = None

    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "property-of-product-samples/": "Endpoint for property-of-product samples for both list and detail view.",
    }


class Trade_samples(FactBaseModel_samples):
    time: int
    product: str
    export_location: str
    import_location: str
    value: float
    unit: str
    flag: Optional[str] = None  # TODO flag rework
    price_type: Optional[str] = None
    source: Optional[str] = None
    account_type: Optional[str] = None
    _classification: ClassVar[Dict[str, Tuple[str, str]]] = {
        "location": ("ISO2", "location"),
        "product": ("bonsai", "flowobject"),
        "activity": ("bonsai", "activitytype"),
    }
    _endpoints: ClassVar[Dict[str, str]] = {
        "trade-samples/": "Endpoint for trade samples for both list and detail view.",
    }



"""Initializes dataio Python package."""

# NOTE: Do not edit here
# Dependencies for __init__ file, d
import sys
from os import listdir
from os.path import dirname
from pathlib import Path

# NOTE: Edit here
# Functions and classes to export
from ._plot import plot

__all__ = ["describe", "validate", "plot", "load", "save", "create"]

STAGE = "util"
EXPORT_PATH = Path(dirname(__file__)).joinpath("_export")
# __all__ = ["datapackage"]

# NOTE: Do not edit from here downward
# Create package version number from git tag
if sys.version_info[:2] >= (3, 8):
    from importlib.metadata import PackageNotFoundError, version
else:
    from importlib_metadata import PackageNotFoundError, version

# Change package version if project is renamed
try:
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError, dist_name, sys

# Remove files and folders starting with underscore from dir()
PATH = Path(dirname(__file__))
for f in listdir(PATH):
    if f[0] == "_":
        STEM = Path(f).stem
        exec(f"from . import {STEM}")
        exec(f"del {STEM}")

del (
    listdir,
    dirname,
    Path,
    PATH,
    STEM,
    f,
)



"""Experiment with delimiter"""

import pandas as pd

from dataio import datapackage as iodp

"""
iodp.describe('data/delimiter/delimiter.dataio.yaml',
              overwrite=True,
              log_name='describe.log')
"""
"""
iodp.validate('data/delimiter/delimiter.dataio.yaml',
              overwrite=True,
              log_name='validate.log')
"""

df = iodp.plot("data/delimiter/delimiter.dataio.yaml")



import datetime
import logging
import os
from decimal import Decimal
from pathlib import Path

import data_io.data_io as io
import pandas as pd
import pytest
import yaml
from data_io.dataclasses import (
    Algorithm,
    AttributeTable,
    Author,
    DataPackage,
    FactTable,
    Field,
    IncompleteDatapackageException,
    License,
    ModifiedBy,
    Source,
)

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def test_file_dir():
    return Path("tests") / "data"


def test_read_basic_data_package_from_disk(test_file_dir):
    """
    This test reads a valid frictionless datapackage to a DataPackage object from the
    current version.
    """

    config = {"name": "datacube", "path": test_file_dir / "test_data_cube"}

    data_package = io.read(config)

    climate_zone_attr_frame = AttributeTable(
        pd.DataFrame(
            [
                ["region with mainly boreal and temperate climate", 0],
                ["region with mainly tropical climate", 1],
            ],
            index=pd.Index(["boreal_temperate", "tropical"], name="code"),
            columns=["description", "position"],
        ),
        {
            "code": Field(name="code"),
            "description": Field(name="description"),
            "position": Field(name="position"),
        },
        "ISO3",
        "misc",
    )
    country_attr_frame = AttributeTable(
        pd.DataFrame(
            [["Algeria", 0], ["Angola", 1], ["Anguilla", 2], ["Argentina", 4]],
            index=pd.Index(["Algeria", "Angola", "Anguilla", "Argentina"], name="code"),
            columns=["description", "position"],
        ),
        {
            "code": Field(name="code"),
            "description": Field(name="description"),
            "position": Field(name="position"),
        },
        "ISO3",
        "region",
    )
    property_attr_frame = AttributeTable(
        pd.DataFrame(
            [
                ["default", "mean", 0],
                ["minimum", "2.5th percentile", 1],
                ["maximum", "97.5th percentile", 2],
            ],
            index=pd.Index(["def", "min", "max"], name="code"),
            columns=["description", "remarks", "position"],
        ),
        {
            "code": Field(name="code"),
            "description": Field(name="description"),
            "remarks": Field(name="remarks"),
            "position": Field(name="position"),
        },
        "ISOSomething",
        "something",
    )
    swds_type_attr_frame = AttributeTable(
        pd.DataFrame(
            [
                ["Managed solid waste disposal site (anaerobic).", 0],
                ["Well managed solid waste disposal site (semi-aerobic).", 1],
                ["Poorly managed solid waste disposal site (semi-aerobic).", 2],
            ],
            index=pd.Index(
                ["managed", "managed_well_s-a", "managed_poorly_s-a"], name="code"
            ),
            columns=["description", "position"],
        ),
        {
            "code": Field(name="code"),
            "description": Field(name="description"),
            "position": Field(name="position"),
        },
        "Something",
        "something",
    )
    unit_attr_frame = AttributeTable(
        pd.DataFrame(
            [
                ["capita", 0],
                ["tonnes per capita per year", 1],
                ["kilogram per kilogram", 2],
                ["per year", 3],
                ["no unit", 4],
            ],
            index=pd.Index(["cap", "t/cap/yr", "kg/kg", "1/yr", "none"], name="code"),
            columns=["description", "position"],
        ),
        {
            "code": Field(name="code"),
            "description": Field(name="description"),
            "position": Field(name="position"),
        },
        "unit",
        "unit",
    )
    year_attr_frame = AttributeTable(
        pd.DataFrame(
            [["year in which activity occurs", 0]],
            index=pd.Index(["2010"], name="code"),
            columns=["description", "position"],
        ),
        {
            "code": Field(name="code"),
            "description": Field(name="description"),
            "position": Field(name="position"),
        },
        "year",
        "period",
    )

    climate_zone_frame = FactTable(
        pd.DataFrame(
            [
                "boreal_temperate",
                "tropical",
                "tropical",
                "boreal_temperate",
            ],
            index=pd.MultiIndex.from_arrays(
                [
                    ["a0", "a1", "a2", "a4"],
                    ["Albania", "Algeria", "Angola", "Argentina"],
                ],
                names=["id", "country"],
            ),
            columns=["value"],
        ),
        fields={
            "value": Field(name="value"),
            "id": Field(name="id", foreign_name="code", resource="climate_zone_dim"),
            "country": Field(name="id", foreign_name="code", resource="country"),
        },
    )
    mcf_frame = FactTable(
        pd.DataFrame(
            [
                1,
                0.9,
                1,
                0.3,
            ],
            index=pd.MultiIndex.from_arrays(
                [
                    ["a0", "a1", "a2", "a4"],
                    ["managed", "managed", "managed", "managed_well_s-a"],
                    ["def", "min", "max", "min"],
                    ["kg/kg", "kg/kg", "kg/kg", "kg/kg"],
                ],
                names=["id", "swds_type", "property", "unit"],
            ),
            columns=["value"],
        ),
        fields={
            "value": Field(name="value"),
            "id": Field(name="id", foreign_name="code", resource="mcf_dim"),
            "swds_type": Field(
                name="swds_type", foreign_name="code", resource="swds_type"
            ),
            "property": Field(
                name="property", foreign_name="code", resource="property_dim"
            ),
        },
    )
    urb_pop_frame = FactTable(
        pd.DataFrame(
            [
                24299170.0,
                13963065.0,
                37055902.0,
            ],
            index=pd.MultiIndex.from_arrays(
                [
                    ["a0", "a1", "a3"],
                    [2010, 2010, 2010],
                    ["Algeria", "Angola", "Argentina"],
                    ["def", "def", "def"],
                    ["cap", "cap", "cap"],
                ],
                names=["id", "year", "country", "property", "unit"],
            ),
            columns=["value"],
        ),
        fields={
            "value": Field(name="value"),
            "id": Field(name="id", foreign_name="code", resource="mcf_dim"),
            "country": Field(name="country", foreign_name="code", resource="country"),
            "property": Field(
                name="property", foreign_name="code", resource="property_dim"
            ),
            "year": Field(name="year", foreign_name="code", resource="year"),
            "unit": Field(name="unit", foreign_name="code", resource="unit"),
        },
    )

    tables = {
        "par_climate_zone": climate_zone_frame,
        "par_mcf": mcf_frame,
        "par_urb_population": urb_pop_frame,
        "dim_climate_zone": climate_zone_attr_frame,
        "country": country_attr_frame,
        "dim_property": property_attr_frame,
        "dim_swds_type": swds_type_attr_frame,
        "unit": unit_attr_frame,
        "year": year_attr_frame,
    }

    exptected_license = License(
        "MIT",
        "MIT License",
        "this is actually not a data license",
        "https://www.mitlicense.com",
    )
    exptected_source = Source(
        reference="Something",
        encoding="utf-8",
        link="https://www.example.com",
        link_doc="http://www.example.com/docs",
        authors=["Joe", "Paul", "Melinda", "Anna"],
        institution="Harvard",
        type="SUT",
        license=exptected_license,
        language="en",
    )
    exptected_author = Author(
        "Valentin Starlinger", "2.-0 LCA Consultants", "valentin.starlinger@lca-net.com"
    )
    exptected_algorithm = Algorithm(name="clean/faostat", version_number=0)
    exptected_modified_by = ModifiedBy(
        authors=[exptected_author], algorithm=exptected_algorithm
    )
    expected_package = DataPackage(
        name="datacube",
        category_path="this/is/something",
        version="1.2.3",
        title="Example of a composite data cube stored in tabular format",
        comment="Based on the IPCC waste data collected by Maik Budzinski",
        tables=tables,
        source=exptected_source,
        modified_by=exptected_modified_by,
        last_modified=datetime.datetime(
            2023, 2, 23, 13, 20, 51, tzinfo=datetime.timezone.utc
        ),
        target_language="en",
        encoding="utf-8",
    )

    assert expected_package.name == data_package.name
    assert expected_package.category_path == data_package.category_path
    assert expected_package.title == data_package.title
    assert expected_package.comment == data_package.comment
    assert expected_package.source == data_package.source
    assert expected_package.modified_by == data_package.modified_by
    assert expected_package.last_modified == data_package.last_modified
    assert expected_package.target_language == data_package.target_language
    assert expected_package.encoding == data_package.encoding

    assert tables.keys() == data_package.tables.keys()

    for name, table in tables.items():
        print(name)
        pd.testing.assert_frame_equal(table.data, data_package.tables[name].data)


@pytest.fixture
def sample_datapackage():
    # Create sample data for testing
    fact_data = pd.DataFrame(
        {"region": [1, 2], "industry": [2, 3], "value": [0.1, 2.3]}
    )
    fact_data = fact_data.set_index(["region", "industry"])
    attr_data = pd.DataFrame({"code": [1, 2], "name": ["Region 1", "Region 2"]})
    attr_data = attr_data.set_index(["code"])
    # Create sample DataPackage instance
    license = License(
        "CC BY 4.0",
        "Creative Commons Attribution 4.0",
        "",
        "https://creativecommons.org/licenses/by/4.0/",
    )
    source = Source(
        "Sample Source", "UTF-8", "", ["John Doe"], "Acme Inc.", "CSV", license, "en"
    )
    author = Author("John Doe", "Acme Inc.", "john.doe@acme.com")
    algorithm = Algorithm("Sample Algorithm", "1.0")
    modified_by = ModifiedBy([author], algorithm)
    tables = {
        "fact_table": FactTable(
            fact_data,
            {
                "region": Field("region", "code", "region"),
                "industry": Field("industry"),
                "value": Field("value"),
            },
        ),
        "region": AttributeTable(
            attr_data,
            {"code": Field("code"), "name": Field("name")},
            "region_name",
            "region",
        ),
    }

    datapackage = DataPackage(
        "sample_data",
        "example/path",
        "1.0.0",
        title="Sample Data",
        comment="Sample data for testing",
        tables=tables,
        source=source,
        modified_by=modified_by,
        last_modified=datetime.datetime(
            2023, 2, 23, 13, 20, 51, tzinfo=datetime.timezone.utc
        ),
        target_language="en",
        encoding="UTF-8",
    )

    return datapackage


def test_write_read(sample_datapackage, tmp_path):
    param_dict = {"root_path": tmp_path}
    io.write(param_dict, sample_datapackage)

    # Test if the required files are created
    data_path = tmp_path / "example" / "path" / "sample_data" / "1" / "0" / "0"
    assert (data_path / "fact_table.csv").exists()
    assert (data_path / "region.csv").exists()
    assert (data_path / "sample_data.datapackage.yaml").exists()

    # print("fact table")
    # with open(data_path / "fact_table.csv", "r") as file:
    #    for line in file.readlines():
    #        print(line, end="")
    # print("attr table")
    # with open(data_path / "region.csv", "r") as file:
    #    for line in file.readlines():
    #        print(line, end="")
    # print("metadata")
    # with open(data_path / "sample_data.datapackage.yaml", "r") as file:
    #    for line in file.readlines():
    #        print(line, end="")

    # Read DataPackage
    param_dict = {"name": "sample_data", "path": data_path}
    read_datapackage = io.read(param_dict)

    print(sample_datapackage)
    print()
    print()
    print(read_datapackage)

    # Compare the original and read DataPackages
    assert sample_datapackage == read_datapackage


def test_incomplete_datapackage_exception(sample_datapackage, tmp_path):
    param_dict = {"root_path": tmp_path}

    # Test missing name
    sample_datapackage.name = None
    with pytest.raises(
        IncompleteDatapackageException, match="DataPackage name field missing"
    ):
        io.write(param_dict, sample_datapackage)
    sample_datapackage.name = "sample_data"

    # Test missing source
    sample_datapackage.source = None
    with pytest.raises(
        IncompleteDatapackageException, match="DataPackage source field missing"
    ):
        io.write(param_dict, sample_datapackage)
    sample_datapackage.source = Source(
        "Sample Source",
        "UTF-8",
        "",
        ["John Doe"],
        "Acme Inc.",
        "CSV",
        sample_datapackage.license,
        "en",
    )

    # Test missing modified_by
    sample_datapackage.modified_by = None
    with pytest.raises(
        IncompleteDatapackageException, match="DataPackage modified_by field missing"
    ):
        io.write(param_dict, sample_datapackage)
    sample_datapackage.modified_by = ModifiedBy(
        [Author("John Doe", "Acme Inc.", "john.doe@acme.com")],
        Algorithm("Sample Algorithm", "1.0"),
    )

    # Test missing target_language
    sample_datapackage.target_language = None
    with pytest.raises(
        IncompleteDatapackageException,
        match="DataPackage target_language field missing",
    ):
        io.write(param_dict, sample_datapackage)
    sample_datapackage.target_language = "en"


def test_version_to_path():
    assert io._version_to_path("1.0.0") == Path("1/0/0")
    assert io._version_to_path("2.1a.3") == Path("2/1a/3")
    assert io._version_to_path("2.3") == Path("2/3")

    with pytest.raises(ValueError):
        io._version_to_path("1")  # Invalid version string

    with pytest.raises(ValueError):
        io._version_to_path("1.a.0.3")  # Invalid version string


def test_category_path_to_path():
    assert io._category_path_to_path("A/B/C") == Path("A/B/C")
    assert io._category_path_to_path("X/Y/Z") == Path("X/Y/Z")

    with pytest.raises(ValueError):
        io._category_path_to_path("A//B/C")  # Invalid category_path string

    with pytest.raises(ValueError):
        io._category_path_to_path("A/B\\C")  # Invalid category_path string

    with pytest.raises(ValueError):
        io._category_path_to_path("A\B\C")  # Invalid category_path string


def test_incomplete_datapackage_exception():
    datapackage = DataPackage("", "", "")

    with pytest.raises(
        IncompleteDatapackageException, match="DataPackage name field missing"
    ):
        io.write({"root_path": "root"}, datapackage)

    datapackage.name = "test"
    with pytest.raises(
        IncompleteDatapackageException, match="DataPackage source field missing"
    ):
        io.write({"root_path": "root"}, datapackage)



from datetime import datetime
from pathlib import Path
from typing import Optional

from dataio.utils.accounts import AccountRepository
from dataio.utils.versions import VersionCollection


class BasePaths:
    def __init__(self, root_path: str | Path):
        if not isinstance(root_path, Path):
            root_path = Path(root_path)
        self.root_path = root_path
        self.bonsai_root = self.root_path / "_bonsai"
        if not self.bonsai_root.exists():
            print("bonsai_root is ", self.bonsai_root)
            self.bonsai_root = self.root_path

            if not self.bonsai_root.exists():
                raise FileNotFoundError(
                    f"Root path <{self.bonsai_root}> is not set correctly. Please reinstantiate"
                )

    @property
    def data_collection(self) -> Path:
        return self.bonsai_root / "collect"

    @property
    def data_clean(self) -> Path:
        return self.bonsai_root / "clean"

    @property
    def exiobase4(self) -> Path:
        from dataio.default import DB_NAME, DB_VERSION

        return self.bonsai_root / "_b Next version" / "Exiobase_4"
        # return self.bonsai_root / "_b Next version" / f"{DB_NAME}_{DB_VERSION}"

    @property
    def merged_collected_data(self) -> Path:
        return self.data_collection / "Merged_data"

    @property
    def correspondence(self) -> Path:
        return self.bonsai_root / "Classification and other data/corrspondence"

    @property
    def classification(self):
        return self.bonsai_root / "Classification and other data/classifications_exiob"


class PathBuilder:
    """
    PathBuilder provides shorthand paths for files on a cloud storage specified by `root_path` to be used by the algorithms
    to create the full exiobase database.

    You can provide a version_source so that you can link to files of previous versions.
    Please resort to `default_version` for the structure of the version file.

    for entries that are not specified in the dictionary, the path to the latest available version is generated.

    Additionally, it is possible to provide the path to an alternative version file that is used instead of the standard one.
    """

    def __init__(
        self,
        root_path: str | Path,
        version_source: Path | str | None = None,
        account_repository: AccountRepository = None,
    ):
        self.base_paths = BasePaths(root_path)
        self.bonsai_home = self.base_paths.root_path
        self.exiobase4 = self.base_paths.exiobase4
        self.correspondence = self.base_paths.correspondence
        self.classification = self.base_paths.classification
        self.data_collection = self.base_paths.data_collection
        self.data_clean = self.base_paths.data_clean
        self.version_source = version_source
        self.versions: VersionCollection = self.load_version_dates(
            source=self.version_source
        )
        self.account_repo = account_repository

    def load_version_dates(
        self,
        source: str | Path | None = None,
    ) -> VersionCollection:
        return VersionCollection.load_from(source)

    def compose(
        self,
        path: Path,
        account_name: Optional[str] = None,
        version_date: Optional[str] = "latest",
    ) -> Path:
        """Ensures a path exists by creating it if necessary.

        This method constructs a path by appending optional account and version date
        components. If the account name is provided, it retrieves the account, adding
        it if it doesn't exist. The version date can be specified or default to the
        latest version. The final path is created if it does not already exist.

            Parameters
            ----------
            path : Path
                The base path to compose.
            account_name : Optional[str], optional
                The account name to include in the path.
            If the account does not exist, it will be added. Defaults to None.
            version_date : Optional[str], optional
                The version date to include in the path.
            It can be "latest" to use the latest version date, a specific date in
            "YYYY-MM-DD" format, or None to exclude the version date. Defaults to "latest".

            Returns
            -------
            Path
                The composed path, created if it didn't already exist.
            Raises
            ------
                KeyError: If the account name is not found in the account repository.
                ValueError: If the provided version date is invalid.
        """
        path_components = [path]
        if account_name:
            try:
                account = self.account_repo.get_account(name=account_name)
            except KeyError:
                print(f"{account_name} not found in accounts")
                account = self.account_repo.add_account(account_name)
            path_components.append(account.name)

            if version_date == "latest":
                version = self.versions.get_latest_version(account.name)
                parsed_version_date = datetime.strptime(
                    version.version_date, "%Y-%m-%d"
                ).strftime("%Y%m%d")
            elif version_date and version_date != "latest":
                try:
                    self.versions.update_account_version_date(
                        account.name, version_date
                    )
                except ValueError:
                    self.versions.add_account_version(
                        account_name=account.name, version_date=version_date
                    )
                parsed_version_date = datetime.strptime(
                    version_date, "%Y-%m-%d"
                ).strftime("%Y%m%d")
            elif version_date is None:
                parsed_version_date = ""

            path_components.append(parsed_version_date)

        elif not account_name:

            if version_date == "latest":
                parsed_version_date = datetime.strptime(
                    version_date, "%Y-%m-%d"
                ).strftime("%Y%m%d")
                path_components.append(parsed_version_date)
            elif version_date and version_date != "latest":
                parsed_version_date = datetime.strptime(
                    version_date, "%Y-%m-%d"
                ).strftime("%Y%m%d")
            elif version_date is None:
                parsed_version_date = ""
            path_components.append(parsed_version_date)
        final_path = Path(*path_components)
        final_path.mkdir(
            parents=True, exist_ok=True
        )  # Create the path if it doesn't exist
        # TODO: append the final path to PathRepository
        return final_path

    def _construct_version(self, account: str) -> str:
        version = self.versions.get_latest_version(account)
        return datetime.strptime(version.version_date, "%Y-%m-%d").strftime("%Y%m%d")

    @property
    def supply_raw(self):
        date_supply = self._construct_version("supply")
        folder_supply = "supply_" + date_supply
        return self.compose(
            self.base_paths.exiobase4 / "Supply" / "Raw" / folder_supply,
            version_date=None,
        )

    @property
    def supply_intermediate(self):
        date_supply_interm = self._construct_version("supply intermediate")
        folder_supply_final = "supply_" + date_supply_interm
        return self.compose(
            self.base_paths.exiobase4 / "Supply" / "Intermediate" / folder_supply_final,
            version_date=None,
        )

    @property
    def balance_raw(self):
        date_supply_interm = self._construct_version("supply intermediate")
        return self.compose(
            self.base_paths.exiobase4 / "Balanced" / date_supply_interm,
            version_date=None,
        )

    @property
    def use_raw(self):
        date_use = self._construct_version("use")
        folder_use_raw = "use_" + date_use
        return self.compose(
            self.base_paths.exiobase4 / "Use" / "Raw" / folder_use_raw,
            version_date=None,
        )

    @property
    def use_intermediate(self):
        date_use_interm = self._construct_version("use intermediate")
        folder_use_final = "use_" + date_use_interm
        return self.compose(
            self.base_paths.exiobase4 / "Use" / "Intermediate" / folder_use_final,
            version_date=None,
        )

    @property
    def hiot_raw(self):
        date_hiot = self._construct_version("hiot")
        folder_hiot_raw = "Raw_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4 / "HIOT" / "Raw" / folder_hiot_raw,
            version_date=None,
        )

    @property
    def hiot_interm(self):
        date_hiot = self._construct_version("hiot")
        return self.compose(
            self.base_paths.exiobase4 / "HIOT" / "Intermediate" / date_hiot,
            version_date=None,
        )

    @property
    def prod_markets(self):
        date_prod_markets = self._construct_version("market of products")
        folder_prod_markets = "product_markets_" + date_prod_markets
        return self.compose(
            self.base_paths.exiobase4 / "Markets" / "Raw" / folder_prod_markets,
            version_date=None,
        )

    @property
    def trade_raw(self):
        date_trade_merged = self._construct_version("trade_merged")
        folder_trade_raw = "raw_data_" + date_trade_merged
        return self.compose(
            self.base_paths.exiobase4 / "Trade" / "Raw" / folder_trade_raw,
            version_date=None,
        )

    @property
    def trade_intermediate(self):
        date_trade_int = self._construct_version("trade intermediate")
        folder_trade_int = "trade_" + date_trade_int
        return self.compose(
            self.base_paths.exiobase4 / "Trade" / "Intermediate" / folder_trade_int,
            version_date=None,
        )

    @property
    def emissions_intermediate(self):
        date_emissions = self._construct_version("emissions")
        folder_emissions = "emissions_" + date_emissions
        return self.compose(
            self.base_paths.exiobase4 / "Emissions" / "Intermediate" / folder_emissions,
            version_date=None,
        )

    @property
    def emissions_raw(self):
        date_emissions = self._construct_version("emissions")
        folder_emissions = "emissions_" + date_emissions
        return self.compose(
            self.base_paths.exiobase4 / "Emissions" / "Raw" / folder_emissions,
            version_date=None,
        )

    @property
    def emissions_coeff(self):
        return self.correspondence.parent / "Emissions_coeffs"

    @property
    def fao_collection(self):
        return self.base_paths.data_collection / "Faostat"

    @property
    def fao_processed(self):
        return self.data_collection / "Faostat" / "processed"

    @property
    def fao_store(self):
        return self.base_paths.data_collection / "Faostat" / "store"

    @property
    def ferts_collection(self):
        return self.base_paths.data_collection / "Fertiliser data"

    @property
    def iLUC_param(self):
        return self.base_paths.data_collection / "iLUC"

    @property
    def dm_coeff(self):
        return self.base_paths.data_collection / "Dry_matter"

    @property
    def iLUC_raw(self):
        date_iluc = self._construct_version("iluc")
        folder_iLUC = "Input_data_" + date_iluc
        return self.compose(
            self.base_paths.exiobase4 / "iLUC" / "Raw" / folder_iLUC,
            version_date=None,
        )

    @property
    def iLUC_interm(self):
        date_iluc = self._construct_version("iluc")
        folder_iLUC = "Derived_data_" + date_iluc
        return self.compose(
            self.base_paths.exiobase4 / "iLUC" / "Intermediate" / folder_iLUC,
            version_date=None,
        )

    @property
    def land_use(self):
        date_land = self._construct_version("land use")
        folder_land_use = "Raw_" + date_land
        return self.compose(
            self.base_paths.exiobase4 / "Land" / "Raw" / folder_land_use,
            version_date=None,
        )

    @property
    def fertilisers(self):
        date_ferts = self._construct_version("fertilisers_prod")
        folder_fers = "Raw_" + date_ferts
        return self.compose(
            self.base_paths.exiobase4 / "Fertilisers" / "Raw" / folder_fers,
            version_date=None,
        )

    @property
    def fert_interm(self):
        date_ferts = self._construct_version("fertilisers_prod")
        folder_fers = "Raw_" + date_ferts
        return self.compose(
            self.base_paths.exiobase4 / "Fertilisers" / "Intermediate" / folder_fers,
            version_date=None,
        )

    @property
    def property_param(self):
        date_property = self._construct_version("property matrix")
        folder_proper = "Raw_" + date_property
        return self.compose(
            self.base_paths.exiobase4 / "Properties" / folder_proper, version_date=None
        )

    @property
    def ipcc_param(self):
        return self.base_paths.data_collection / "IPCC"

    @property
    def gams_inputs(self):
        return self.base_paths.data_collection / "GAMS" / "data_source_for_GAMS"

    @property
    def corresp_fao(self):
        return self.correspondence / "fao"

    @property
    def hiot_with_iluc(self):
        date_hiot = self._construct_version("hiot")
        folder_iluc_out = "source_data_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4 / "HIOT" / "HIOT_with_iluc" / folder_iluc_out,
            version_date=None,
        )

    @property
    def hiot_with_marg_elect(self):
        date_hiot = self._construct_version("hiot")
        folder_elec_out = "source_data_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4
            / "HIOT"
            / "HIOT_with_marg_electricity"
            / folder_elec_out,
            version_date=None,
        )

    @property
    def hiot_with_capital(self):
        date_hiot = self._construct_version("hiot")
        folder_capital_out = "source_data_" + date_hiot

        return self.compose(
            self.base_paths.exiobase4 / "HIOT" / "b2_version" / folder_capital_out,
            version_date=None,
        )

    @property
    def value_added(self):
        date_hiot = self._construct_version("hiot")
        folder_value_added = "source_data_" + date_hiot

        return self.compose(
            self.base_paths.exiobase4 / "HIOT" / "value_added" / folder_value_added,
            version_date=None,
        )

    @property
    def matrix_of_invest(self):
        date_hiot = self._construct_version("hiot")
        folder_capital_mat = "source_data_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4
            / "HIOT"
            / "b2_version"
            / "matrix_investments"
            / folder_capital_mat,
            version_date=None,
        )

    @property
    def iea_clean(self):
        return self.data_clean / "IEA"

    @property
    def iea_raw_exio(self):
        date_iea = self._construct_version("IEA")
        folder_iea = "raw_data_" + date_iea
        return self.compose(
            self.base_paths.exiobase4 / "IEA" / "Raw" / folder_iea, version_date=None
        )

    @property
    def iea_interm(self):
        date_iea = self._construct_version("IEA")
        folder_iea_interm = "iea_data_" + date_iea
        return self.compose(
            self.base_paths.exiobase4 / "IEA" / "Intermediate" / folder_iea_interm,
            version_date=None,
        )

    @property
    def lci_raw(self):
        return self.base_paths.data_collection / "LCI"

    @property
    def lci_cleaned(self):
        return self.compose(self.data_clean, account_name="LCI", version_date=None)

    @property
    def lci_vehicles(self):
        return (
            self.base_paths.data_collection
            / "Product_statistics"
            / "Vehicles"
            / "ready_for_exiobase"
        )

    @property
    def lci_products(self):
        return self.base_paths.data_collection / "Product_statistics"

    @property
    def lci_concito(self):
        return self.base_paths.data_collection / "Concito_DB"

    @property
    def lci_exio4(self):
        date_use = self._construct_version("use")
        folder_lci = "Raw_" + date_use
        return self.compose(
            self.base_paths.exiobase4 / "Product_statistics" / folder_lci,
            version_date=None,
        )

    @property
    def heat_markets(self):
        date_heatmarket = self._construct_version("heatmarket")
        folder_heatmarket = "heat_markets_" + date_heatmarket
        return self.compose(
            self.base_paths.exiobase4 / "Markets" / "Raw" / folder_heatmarket,
            version_date=None,
        )

    @property
    def fish_markets(self):
        date_fish = self._construct_version("fish prod mix")
        folder_fishmarket = "fish_prod_mix_" + date_fish
        return self.compose(
            self.base_paths.exiobase4 / "Markets" / "Raw" / folder_fishmarket,
            version_date=None,
        )

    @property
    def waste_accounts(self):
        date_waste = self._construct_version("waste supply")
        folder_waste = "Raw_" + date_waste
        return self.compose(
            self.base_paths.exiobase4 / "Waste" / "Raw" / folder_waste,
            version_date=None,
        )

    @property
    def waste_markets(self):
        date_wastemarkets = self._construct_version("waste markets")
        folder_waste_markets = "waste_markets_" + date_wastemarkets
        return self.compose(
            self.base_paths.exiobase4 / "Markets" / "Raw" / folder_waste_markets,
            version_date=None,
        )

    @property
    def monetary_tables(self):
        return self.base_paths.data_collection / "MRSUT_2016"

    @property
    def property_values(self):
        date_property = self._construct_version("property matrix")
        folder_property = "Raw_" + date_property
        return self.compose(
            self.base_paths.exiobase4 / "Properties" / folder_property,
            version_date=None,
        )

    @property
    def prices(self):
        return self.compose(
            self.base_paths.exiobase4,
            account_name="prices",
            version_date="latest",
        )

    @property
    def cleaned_exio_3(self):
        return self.compose(
            self.base_paths.data_clean / "Exiobase_v3", version_date=None
        )

    @property
    def un_data(self):
        return self.compose(
            self.base_paths.data_collection / "UN_commodities", version_date=None
        )

    @property
    def un_data_elab(self):
        date_un = self._construct_version("UNcd")
        folder_un = "intermediate_" + date_un
        return self.compose(
            self.base_paths.exiobase4 / "UN_data" / folder_un, version_date=None
        )

    @property
    def b2_version(self):
        date_hiot = self._construct_version("hiot")
        folder_b2_version = "data_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4
            / "HIOT"
            / "b2_simapro_version"
            / folder_b2_version,
            version_date=None,
        )

    @property
    def outlier(self):
        date_hiot = self._construct_version("hiot")
        folder_outlier = "version_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4 / "analyses" / "outlier_analysis" / folder_outlier
        )

    @property
    def simapro(self) -> Path:
        return self.compose(self.base_paths.exiobase4 / "SimaPro")

    @property
    def cement(self):
        # TODO: Why cement have the same version as hiot?
        date_hiot = self._construct_version("hiot")
        folder_cement = "version_" + date_hiot
        return self.compose(
            self.base_paths.exiobase4
            / "parameterized_functions"
            / "cement"
            / folder_cement
        )

    @property
    def cement_data(self):
        return self.compose(
            self.base_paths.exiobase4 / "parameterized_functions" / "cement_data"
        )

    @property
    def natural_resource(self) -> Path:
        date_raw = self._construct_version("natural_resource")
        folder_res = "raw_data_" + date_raw
        return self.compose(
            self.base_paths.exiobase4 / "Resource_extraction" / folder_res,
            version_date=None,
        )

    @property
    def cleaned_fao(self) -> Path:
        return self.data_clean / "Faostat" / "new format"

    @property
    def cleaned_forestry(self):
        date_forestry = self._construct_version("forestry")
        folder_fores = "cleaned_data_" + date_forestry
        return self.compose(
            self.base_paths.exiobase4 / "Faostat" / "forestry" / folder_fores,
            version_date=None,
        )

    @property
    def lci_fish(self):
        date_lci_fish = self._construct_version("lci_fish")
        folder_lci_fish = "raw_data_" + date_lci_fish
        return self.compose(
            self.base_paths.exiobase4 / "LCI" / "aquaculture" / folder_lci_fish,
            version_date=None,
        )

    @property
    def fao_raw(self):
        date_lci_fish = self._construct_version("lci_fish")
        folder_fao = "raw_data_" + date_lci_fish
        return self.compose(
            self.base_paths.exiobase4 / "Faostat" / folder_fao, version_date=None
        )

    @property
    def lci_country(self):
        return self.exiobase4 / "LCI" / "country_activ_specific"

    @property
    def lci_act_generic(self):
        return self.exiobase4 / "LCI" / "default_act_specific"

    @property
    def lci_prod_generic(self):
        return self.exiobase4 / "LCI" / "default_prod_specific"

    @property
    def trade_route(self):
        date_trade_route = self._construct_version("trade_route")
        folder_trade_route = "trade_route_" + date_trade_route
        return self.compose(
            self.base_paths.exiobase4 / "trade_route" / "raw" / folder_trade_route,
            version_date=None,
        )

    def list_path_attributes(self):
        import inspect

        paths = []
        for name, _ in inspect.getmembers(
            self.__class__, lambda v: isinstance(v, property)
        ):
            paths.append(name)
        return paths
